const LINGO_TYPE_ANNE = 0;
const LINGO_TYPE_GRAPH = 1;
const LINGO_TYPE_COMMON = 2;

const TENSE_PRESENT = 1;
const TENSE_PAST = 2;
const TENSE_PAST_COMPLETE = 3;

const _BLACK = 30;
const _RED = 31;
const _GREEN = 32;
const _YELLOW = 33;
const _YELLOW_i = 330;
const _BLUE = 34;
const _BLUE_i = 340;
const _PINK = 35;
const _CYAN = 36;
const _WHITE = 37;

//This is not semantic langauge via clumps.  This is just a basic output
//class for debugging or quick english display of things.
class anne_talk extends anne_singleton {
	constructor(){
		super();
		
		//THIS IS A SIMPLE WAY TO PROGRAM GRAMMAR utilizing base roles as storage.
		//different languages can have different grammar patterns.
		//one could spend a very long time in here building a robust langauge handler..
		//it can be done for AI/AGI use cases.
		
		//WE CAN STORE THE ROLES PROPERLY.
		//WE CAN VERBALIZE THEM.
		//In time, (with a fuller vocab) we can fully parse and conceptualize into the ANNE semantics.

		//ANNE training is done by experience/use..  The very basics are put in at inception to have something 
		//to work with.  The idea is parts-of-speachless (we use a higher concept of part of anne)
		//Language is a map into a higher-order brain protocol.
		//The Language may change, and with it the ways to communicate, but the concepts are stored using the 
		//higher-order unified brain protocol.  The same brain can speak 4 different languages and doesn't rqeuire
		//4 unique concepts repeating what a "dog" is defined as.  There are 4 differeent lang word maps to the 
		//same concepts (if it is indeed the same concept) 

		//clump definition pattern.
		this.def = new Object();
		this.def.infini = {pat_part_name:"infini", roles:[{role:"role_action", p:"to"}]};	//no use ot do doInfinitive w/ rule as it is just a "to"+
		this.def.mode = {pat_part_name:"mode", roles:["role_mode"]};
		this.def.obj = {pat_part_name:"obj", roles:[{role:'role_affected_thing', mod:'role_affected_thing_mod', det_func:this.___doThingDet}, 
													{role:'role_resulting_new_thing', mod:'role_resulting_new_thing_mod', det_func:this.___doThingDet},
													{role:'role_theme_thing', mod:'role_theme_thing_mod', det_func:this.___doThingDet},
													'role_instrument_mod',
													{role:'role_instrument', mod:'role_theme_thing_mod', det_func:this.___doThingDet},
													'role_state', 'role_feelz', 'role_thought', 'role_fact']};

		this.def.manner = {pat_part_name:"manner", roles:[{role:"role_manner", modfunc:this.___doMannerEngShifter}]}
													
		this.def.wherewhen2 = {
			pat_part_name:"wherewhen2",
			roles:['role_source', 'role_path', 'role_goal',
				{role:'role_where', mod:'role_where_mod', p_default:"to"}, 
				{role:'role_when', mod:'role_when_mod'},
				'role_start','role_end']
		};

		this.eng_pat_def = [
			this.def.infini,
			this.def.mode,
			this.def.obj,
			this.def.manner,
			this.def.wherewhen2
		]


		//action inst pattern
		this.actionInst = new Object();
		this.actionInst.subj = {pat_part_name:"subj", roles:['role_actor']};
		this.actionInst.mode = {pat_part_name:"mode", roles:['role_manner', 'role_mode']};
		this.actionInst.pred = {pat_part_name:"pred", roles:['role_action'], mod:'role_tense', modfunc:this.englishTensify}; //'role_reason'
		this.actionInst.obj = {pat_part_name:"obj", roles:['role_affected_thing', 'role_resulting_new_thing', 
										'role_state', 'role_feelz', 'role_thought', 'role_fact',
										'role_instrument_mod','role_instrument']};

		this.actionInst.theme_thing = {pat_part_name:"theme_thing", roles:['role_theme_thing_mod', 'role_theme_thing']};
		
		//to use role_source/path/goal in the same clump you will likely want 
		//more than one role_where_mod.  if so, tuff beans.  Create a nested neuron.
		this.actionInst.wherewhen = {
				pat_part_name:"wherewhen",
				roles:['role_source', 'role_path', 'role_goal',
				'role_where_mod', 'role_where', 'role_when_mod', 'role_when', 
				'role_start', 'role_end']
			};
		this.actionInst.wherewhen2 = {
			pat_part_name:"wherewhen2",
			roles:['role_source', 'role_path', 'role_goal',
				{role:'role_where', mod:'role_where_mod', p_default:"at", det_func:this.___doWhereDet}, 
				{role:'role_when', mod:'role_when_mod', p:"on"}
				   ,'role_start', 'role_end']
		}
		this.actionInst.pre_connectors = {
			pat_part_name:"pre_connectors",
			recurse:true,
			roles:['role_cause', 'role_cdx', 'role_prev']};
		this.actionInst.post_connectors = {
			recurse:true,
			pat_part_name:"post_connectors",
			roles:['role_effect', 'role_result', 'role_next']};

		//this.actionInst.tense,
		this.eng_pat1 = [
			this.actionInst.subj, 
			this.actionInst.mode,
			this.actionInst.pred, 
			this.actionInst.theme_thing, 
			this.actionInst.obj, 
			this.actionInst.wherewhen2,
			this.actionInst.pre_connectors,
			this.actionInst.post_connectors
		];


		this.type_pre = ["sys", "be", "param", "has", "awareness", "exper", "clump", "sys msg", "brain", "role", "episodic"];

		this.vowel=["a", "e", "i", "o", "u"];  //and sometimes y? lolz.
		
		this.init();
	}


	init = function() {
		this.bgbigred = this.redbig;
		this.bgblack = 'background:black; color:white';
		this.bgblackbold = 'background:black; color:white; font-weight:bold';
		this.bgblackyellow = 'background:#171717; color: #fefe17';
		this.bgblue = 'background:blue; color: white';
		this.bgbluebig = 'background:blue; color: white; font-size:1.1rem';
		this.bgerror = 'background: red; color: white; font-weight:bold; font-size:1.1rem;';
		this.bgfire = 'background:#F7C24A; color: #171717';
		this.bggray = 'background:#ccc; color: #171717';
		this.bggreen = 'background: green; color: white';
		this.bggreenbig = 'background: green; color: white; font-size:1.1rem;';
		this.bgivoryblue = 'background:ivory; color: blue';
		this.bglightblue = 'background:lightblue; color: #171717';
		this.bglightbluebold = 'background:lightblue; color: #171717; font-weight:bold';
		this.bglightgreen = 'background:lightgreen; color: #171717';
		this.bglightred = 'background:lightgreen; color: #171717';
		this.bgneg = this.bgred;
		this.bgneu = this.bgyellow;
		this.bgorange = 'background: orange; color: #171717';
		this.bgpink = 'background:pink; color: #171717';
		this.bgcyan = 'background:cyan; color: #171717';
		this.bgpos = 'background: #00ff00; color: #171717';
		this.bgpurple = 'background:purple; color: white';
		this.bgred = 'background: red; color: white';
		this.bgredbig = 'background: red; color: white; font-size:1.1rem;';
		this.bgyellow = 'background:#fefe17; color: black';
		
		this.bgheading_yellow = 'background:black; color: #fefe17; font-size:1.1rem';
		this.bgheading_black = 'background:whitesmoke; color: black; font-size:1.1rem';
		this.bgheading = 'background:ivory; color: #171717; font-size:1.4; font-weight:bold';

		this.fromMsg = 'background:ivory; color: #171717; font-weight:bold';
		this.funct = 'background:pink; color: #171717; font-style:italic';
		this.ghosts = 'background:white; color: blue';
		this.msg = 'background: green; color: white';
		this.msgNote = 'background:ivory; color:green; font-style:italic';
		this.pooltitle = 'background:ivory; color: black; font-weigt:bold; font-size:1.2rem;';
		this.title = this.pooltitle;
		
		this.bold_black = 'background:ivory; color: #171717; font-weight:bold';
		this.black = 'background:ivory; color: #171717';
		this.green = 'background: ivory; color: green';
		this.orange = 'background: ivory; color: orange;';
		this.redbig = 'background: red; color: white';
		this.yellow = 'background: ivory; color: yellow';
		
		this.part_cnt = this.bgpink;
		this.part_depth = this.bgyellow;
		this.part_id = this.bglightblue;
		this.part_full = this.bggreen;
		this.part_hex = this.bglightblue;
		this.part_mod = this.bglightred;
		this.part_mod_ph = this.bgorange;
		this.part_part = this.bglightgreen;
		this.part_sig = this.bglightblue;
		
		this.action = 'background:red;color: white; border-width:1px. border-style:solid; border-color:whitesmoke';
		this.aux = 'background:salmon; color: white; border-width:1px. border-style:solid; border-color:whitesmoke';
		this.connector = 'background:purple; color: white; border-width:1px. border-style:solid; border-color:whitesmoke';
		this.manner = 'background:indianred; color: white; border-width:1px. border-style:solid; border-color:whitesmoke';
		this.mode = 'background:firebrick; color: white; border-width:1px. border-style:solid; border-color:whitesmoke';
			
		this.tab0 ='';
		this.tab1 ='██';
		this.tab2 ='████';
		this.tab3 ='██████';
		this.tab4 ='████████';
		this.tab5 ='██████████';
		this.tab6 ='████████████';
		this.tab7 ='██████████████';
		this.tab8 ='████████████████';
		this.tab9 ='██████████████████';
		this.tab10='████████████████████';
	}


	printf=function(...$args) {
		var lastColor = _WHITE;
		var html = "";
		$args.forEach($arg => {
			if(typeof $arg==="string") html+="<span class='copr-"+lastColor+"'>"+$arg+"&nbsp;</span>";
			else {
				lastColor = parseInt($arg);
				if (lastColor==undefined) lastColor = _WHITE;	
			}
		});
		return html;
	}

	console_color_map = function($int) {
		if ($int==_BLACK) return this.bgblack;
		else if ($int==_RED) return this.bgred;
		else if ($int==_GREEN) return this.bggreen;
		else if ($int==_YELLOW) return this.bgyellow;
		else if ($int==_BLUE) return this.bgblue;
		else if ($int==_PINK) return this.bgpink;
		else if ($int==_CYAN) return this.bgcyan;
		else if ($int==_WHITE) return this.bgwhite;
		return this.bgblack;
	}
	
	printf_console=function(...$args) {
		var lastColor = this.bgblack;
		var strs = [];
		var colors = [];
		$args.forEach($arg => {
			if(typeof $arg==="string") {
				strs.push("%c"+$arg);
				colors.push(lastColor);
			}
			else {
				lastColor = this.console_color_map(parseInt($arg));
				if (lastColor==undefined) lastColor = this.bgwhite;	
			}
		});
		console.log(strs.join(""),...colors);
	}



	/**
	 * 
	 * @param {*} $vb 
	 * @param {*} $tense 
	 * @returns 
	 */
	 englishTensify = function($vb, $tense) {
		if ($tense=="past tense") {
			var rev = $vb.split("").reverse().join("");
			var c0 = rev[0];
			var c1 = "";
			if (rev.length>1) c1 = rev[1];
			if (c0=="e") return $vb+"d";
			if (c0=="y" && $text.length>1) return $vb.substring(0, $vb.length-1)+"ied";
			if (this.vowel.indexOf(c1)!=-1 && this.vowel.indexOf(c0)==-1) return $vb+c0+"ed";
			if (c0=="c") return $vb+"ked"; 
			return $vb+"ed";
		}
		else return $vb;
	}


	/**
	 * For a given role, what is the EC type we expect to fill this role?
	 * @param {String} $role_nid - nid long string 
	 * @returns 
	 */
	__getTopMatchRoleFromRole=function($role_nid) {
        if ($role_nid==MEM.role_action_thing.temp_nid) return MEM.ec_thing.temp_nid;
        else if ($role_nid==MEM.role_theme_thing.temp_nid) return MEM.ec_thing.temp_nid;
        else if ($role_nid==MEM.role_action.temp_nid) return MEM.ec_action.temp_nid;
        else if ($role_nid==MEM.role_thing.temp_nid) return MEM.ec_thing.temp_nid;
        else if ($role_nid==MEM.role_actor.temp_nid) return MEM.ec_thing.temp_nid;
        else if ($role_nid==MEM.role_affected_thing.temp_nid) return MEM.ec_thing.temp_nid;
        else if ($role_nid==MEM.role_posn_thing.temp_nid) return MEM.ec_thing.temp_nid;
        else if ($role_nid==MEM.role_be_thing.temp_nid) return MEM.ec_thing.temp_nid;
        else if ($role_nid==MEM.role_state.temp_nid) return MEM.ec_state.temp_nid;
        else if ($role_nid==MEM.role_when.temp_nid) return MEM.ec_thing.temp_nid;
        else if ($role_nid==MEM.role_where.temp_nid) return MEM.ec_thing.temp_nid;
        else if ($role_nid==MEM.role_manner.temp_nid) return MEM.ec_manner.temp_nid;
        else if ($role_nid==MEM.role_mode.temp_nid) return MEM.ec_mode.temp_nid;
        else if ($role_nid==MEM.role_connector.temp_nid) return MEM.ec_expect_more_of_prev.temp_nid;
        else if ($role_nid==MEM.role_aux_action.temp_nid) return MEM.ec_expect_action.temp_nid;
        else if ($role_nid==MEM.role_thing_mod.temp_nid) return EM.ec_expect_thing_mod.temp_nid;
        else if ($role_nid==MEM.role_question.temp_nid) return MEM.role_question.temp_nid;
        else {
            var nOBJ = MEMG.getNeuronByTempNid($temp_nid);
            console.log("%c unhandled (__getTopMatchRoleFromRole) %c "+nOBJ.text, TALK.bgred, TALK.green);
        }
    }

	/**
	 * 
	 * @param {*} $balanceObj 
	 */
	printBalanceObj=function($balanceObj) {
		console.log($balanceObj);
		console.log("%c annecoin "+ANNE.UTILS.commas($balanceObj.annecoin), TALK.bgblackyellow);
		console.log("%c annies "+ANNE.UTILS.commas($balanceObj.annies), TALK.bglightblue);

		if ($balanceObj.avail_now_annies!=undefined) {
			console.log("%c anncoin available now "+ANNE.UTILS.commas($balanceObj.avail_now_annies/100), TALK.bglightblue);
		}
		if ($balanceObj.avail_soon_annies!=undefined) {
			console.log("%c anncoin available now "+ANNE.UTILS.commas($balanceObj.avail_soon_annies/100), TALK.bglightblue);
		}
	}

	/**
	 * Shift an english word from the stored EC type to the desired EC type.
	 * 
	 * @param {anne_neuron} $roleNOBJ - the role we are shifting from. 
	 * @param {*} $valueNOBJ - the value of the role we are shiting from.
	 * @returns 
	 */
	__engShifter = function($roleNOBJ, $valueNOBJ) {
		console.log("(TALK.__engShifter)", $roleNOBJ, $valueNOBJ);

		//the top EC type of the role we are shifting from.
		var ec_top_desired_tnid = ANNE.TALK.__getTopMatchRoleFromRole($roleNOBJ.temp_nid);
		if (ec_top_desired_tnid==null) return;

		//the ec type we want to shift to.
		var ec_top_observed_tnid = $valueNOBJ.setECConceptualizingParent();
			
		//if the type is already right, return the text of the from.  No shift is needed.
		if (ec_top_desired_tnid==ec_top_observed_tnid) return $valueNOBJ.text;
		
		//we need to find the rule that shifts this concept from ec_top_observed_nid to the ec_top_desired_nid!
		for (var g=0; g < MEM.eng_grammar_rule.children.length; g++) {
			var ruleNOBJ = MEMG.getNeuronByTempNid( MEM.eng_grammar_rule.children[g]);
			var role_src = MEMG.getRoleRelon(ruleNOBJ.outs, MEM.role_source.temp_nid);
			if (role_src==null) continue;
			if (role_src.t!=ec_top_observed_tnid) continue;	

			//we have a rule that starts from where our concept starts.
			//now we need to see if it converts to our desired
			var role_goal = MEMG.getRoleRelon(ruleNOBJ.outs, MEM.role_goal.temp_nid);
			if (role_goal==null) continue;
			if (role_goal.t!=ec_top_desired_tnid) continue;

			//we found a rule!
			//we need to now look for the suffixes that use this rule.
			for (var i=0; i < ruleNOBJ.ins.length; i++) {
				var in_rel = ruleNOBJ.ins[i];
				if (in_rel.type!=TYPE_HAS) continue;
				if (in_rel.r!=MEM.eng_grammar_rule.temp_nid) continue;

				//we found a suffix that uses this rule.
				var suffixNOBJ = MEMG.getNeuronByTempNid(in_rel.f);
				if (suffixNOBJ==null) continue;
				return TALK.__applyOutEngSuffix(suffixNOBJ, $valueNOBJ);
			}
		}

		console.log("%c could not find rule.");
	}


	/**
	 * 
	 * @param {anne_neuron} $suffixNOBJ 
	 * @param {anne_neuron} $valueNOBJ 
	 * @returns 
	 */
	__applyOutEngSuffix=function($suffixNOBJ, $valueNOBJ) {
		console.log("(TALK.__applyOutEngSuffix)");
		var suffixOBJ  = ANNE.BRAIN.NLP.LC._suffixMap.getByText($suffixNOBJ.text);
		if (suffixOBJ==null || suffixOBJ.cdx_rule_out==undefined) {
			//console.log("%c NO KNOWN OUTBOUND ENGLISH RULE FOR SUFFIX "+$suffixNOBJ.text, TALK.bgredbig);
			return $valueNOBJ.text+$suffixNOBJ.text;
		}
		return $suffixNOBJ.cdx_rule_out($valueNOBJ.text);
	}



	/**
	 * Take a manner role or roles and convert them to be proper adverbs if they are not adverbs.
	 * @param {Array} $manner_roles 
	 * @returns 
	 */
	___doMannerEngShifter = function($manner_roles) {
		console.log("(TALK.___doMannerEngShifter)", $manner_roles);
		var manner_text = [];
		for (var w=0; w < $manner_roles.length; w++) {
			var rel_manner = $manner_roles[w];
			var toNOBJ = MEMG.getNeuronByTempNid(rel_manner.t);
			var manner = ANNE.TALK.__engShifter(MEM.role_manner, toNOBJ);
			manner_text.push(manner);
		}
		return TALK.listifyStrArr(manner_text);
	}

	___doThingDet = function($thing_roles) {
		console.log("(TALK.___doWhereDet)", $thing_roles);
		var thing_text = [];
		for (var w=0; w < $thing_roles.length; w++) {
			var rel_where = $thing_roles[w];
			var toNObj = MEMG.getNeuronByTempNid(rel_where.t);
			if (toNObj.ntype==NTYPE_CLASS) {
				thing_text.push(ANNE.TALK.getDet(toNObj.text)+" "+toNObj.text);
			}
			else {
				if (toNObj.bes.indexOf(MEM.site.temp_nid)!=-1 && toNObj.text.toLowerCase().indexOf("the")!=0) {
					thing_text.push("the "+toNObj.text);
				}
				else thing_text.push(toNObj.text);
			}
		}
		return TALK.listifyStrArr(thing_text);
	}

	___doWhereDet = function($where_roles) {
		//console.log("(TALK.___doWhereDet)", $where_roles);
		var where_text = [];
		for (var w=0; w < $where_roles.length; w++) {
			var rel_where = $where_roles[w];
			var toNObj = MEMG.getNeuronByTempNid(rel_where.t);
			if (toNObj.ntype==NTYPE_CLASS) {
				where_text.push(ANNE.TALK.getDet(toNObj.text)+" "+toNObj.text);
			}
			else {
				if (toNObj.bes.indexOf(MEM.site.temp_nid)!=-1 && toNObj.text.toLowerCase().indexOf("the")!=0) {
					where_text.push("the "+toNObj.text);
				}
				else where_text.push(toNObj.text);
			}
		}
		return TALK.listifyStrArr(where_text);
	}

	/**
	 * This print out the meaning (infinitive form definition) of an action neuron.
	 * Whatever ROLES are stored in the clump will define the clump.
	 * 
	 * We can also show connotations via TYPE_BRAIN and assocs but it will be 
	 * better to do that in the dossier file for this concept where we can display them
	 * more cleanly.
	 * 
	 * @param {anne_neuron} $clumpNOBJ 
	 * @param {Object} $gpat 
	 * @param {Boolean} $bDEBUG 
	 * @returns 
	 */
	verbalizeClassActionDef = function($clumpNOBJ, $gpat=null, $bDEBUG=true, $tenseOverride="", $lvl=0) {
		if  ($gpat==null) $gpat = this.eng_pat_def;
		if ($clumpNOBJ==null) return;
		var output = "";
		var parts = [];
		var order = [];

        if ($bDEBUG==true) {
			console.log("%c (verbalizeClassActionDef) lvl:"+$lvl+" for clump"+$clumpNOBJ.text, TALK.bggreenbig);
		}

		var part_of_anne = null;
		var role_pre_var = "";

		//if this is a class and not present tense, rewire to verbalize the parent action instead.
		if ($clumpNOBJ.ntype==NTYPE_CLASS && $clumpNOBJ.tense!="present tense" && $tenseOverride=="") {
			role_pre_var = "_meaning_";
		}
		
		console.log("role_pre_var: "+role_pre_var);

		//will have to fix roles here for EC TYPE shifts.. hmm.
		var doModNest=function($mods, $lvl=0, $det_func=null, $bDETFIRSTONLY=false) {
			console.log("(doModNest) lvl"+$lvl);
			var modTxt = "";
			var keys = Object.keys($mods)
			for (var k=0; k < keys.length; k++) {
				var rObj = $mods[ keys[k] ];
				if (rObj.mods!=undefined) {
					modTxt += doModNest(rObj.mods, $lvl+1, $det_func, true)+" ";
					$det_func = null;
				}
				if (rObj.nameless!=undefined) continue;
				var this_text =  "";
				if (rObj.r!=MEM.role_connector.temp_nid && $det_func!=null) {
					if ($bDETFIRSTONLY==true && k==0) this_text = $det_func([rObj]);
					else this_text = MEMG.__MASTER_DISPLAY_RELON_TO(rObj);
				} 
				else this_text = MEMG.__MASTER_DISPLAY_RELON_TO(rObj);
				modTxt+=this_text+" ";
			}
			return modTxt;
		}

		for (var p = 0; p < $gpat.length; p++) {
            var part = $gpat[p];
            var roles = part.roles;
			if ($bDEBUG==true) console.log("%c"+part.pat_part_name, TALK.bgblack);

			//loop the roles in the part roles array.
			//if the role is a string, it is just simple.
			//if it is an object, we have .p, .mod, .modfunct, and .role props.
            for (var r=0; r < roles.length; r++) {
                var role = roles[r];
				var pre = "";
				var pre_default = "";
				var mod = "";
				var modfunc = null;
				var det_func = null;
				if (typeof role !=="string") {
					pre = role.p;
					if (pre==undefined) pre = "";
					mod = role.mod;
					if (mod==undefined) mod = null;
					modfunc = role.modfunc;
					if (modfunc==undefined) modfunc =null;
					det_func = role.det_func;
					if (det_func==undefined) det_func = null;
					pre_default = role.p_default;
					if (pre_default == undefined) pre_default = null;
					else {
						if (mod==null) pre = pre_default;	
					}
					role = role.role;	//set this last or we lose our others.
				}

				var arr = $clumpNOBJ[role_pre_var+role];
                if (ANNE.UTILS.isset(arr)==false) continue;

				var mod_arr = null;
				if (mod!=null) {
					mod_arr = $clumpNOBJ[role_pre_var+mod];
					if (ANNE.UTILS.isset(mod_arr)==false) {
						mod=null;
						mod_arr=null;
					}
				}

                var to = "";
				order = order.concat(arr);


				//if this is a role_action.. and the the tense is present.
				if (role=="role_action") {
					var actionNOBJ = MEMG.getNeuronByNid(arr[0].to_s);
					if ($clumpNOBJ.shift_lang==true) {

						//yet we have a role_tense shifter..  apply the tense.
						to = TALK.englishTensify(actionNOBJ.text, $clumpNOBJ.tense);
					}

					else to = TALK.verbalizeRolesArr(arr);

					part_of_anne = actionNOBJ.getECTop();
				}

				//mod mods the WHOLE phrase if there are multiples.
				else if (role=="role_manner" && modfunc!=null) {
					to = modfunc(arr);
				}

				//{role:'role_where', mod:'role_where_mod', p:"to"},
				else if (modfunc!=null) {
					to = modfunc(arr, mod);
				}


				//new method of mods..
				else if (arr[0].mods!=undefined) {

					console.log("mods in role "+role, arr[0].mods);
					mod_arr = arr[0].mods;
					var modTxt = doModNest(mod_arr, 0, det_func);

					//the base role that has these mods (in eng pat we put the mod BEFORE)

					//note, this should not be an array, only one with our mods method.
					//will pinch it off here -- (will write a single verbalizeRole in a bit)
					to = modTxt+TALK.verbalizeRolesArr([ arr[0] ]);	

					//if a sig, don't add the default if we have it
					if (pre_default!=null && (role=="role_where" || role=="role_when")) {
						var sig = ANNE.BRAIN.NLP.LC._getPhSignals(mod_arr[0].role_text);
						if (sig==null) pre = pre_default;
						else if (sig.mod !=true) pre = pre_default; 
					}
				}

				else if (mod_arr!=null) {
					var newmod = TALK.verbalizeRolesArr(mod_arr);
					to = newmod+" "+TALK.verbalizeRolesArr(arr);

					//if a sig, don't add the default if we have it
					if (pre_default!=null && (role=="role_where" || role=="role_when")) {
						//var modNOBJ = MEMG.getNeuronByTempNid(mod_arr[0].t); //should also be role_text already..
						//console.log(modNOBJ);
						var sig = ANNE.BRAIN.NLP.LC._getPhSignals(mod_arr[0].role_text);
						if (sig==null) pre = pre_default;
						else if (sig.mod !=true) pre = pre_default; 
					}
				}

				else {
					if (det_func!=null) to = det_func(arr);
					else to = TALK.verbalizeRolesArr(arr);
				}
				

				if (to=="") continue;	//do period still. whatever.
				if (pre!="" && pre!=undefined && pre!=null && pre.length>0) {
					parts.push({pre:pre, str:to, output:pre+" "+to, role:role, roles:arr});
					to = pre+" "+to;
				}
				else parts.push({pre:pre, str:to, output:to, role:role, roles:arr});
			
				if ($bDEBUG==true) console.log("%c"+role+" %cpre:"+pre+" %cto:"+to, TALK.bgblackyellow, TALK.bgred, TALK.bggreen);
			}
        }

		for (var p=0; p < parts.length; p++) {
			if (p == (parts.length-1)) output=output+parts[p].output+".";
			else output=output+parts[p].output+" ";
		}

		console.log(output);
		$clumpNOBJ.eval = output;
		if ($lvl!=0) return output;

		if ($bDEBUG==true) console.log("%c"+output, TALK.bggreen);

		var tense = $clumpNOBJ.tense;
		if ($tenseOverride!="") tense = $tenseOverride;

		return new anne_babytalk(output, order, part_of_anne, $clumpNOBJ.tense);
	}

	

    /**
     * Rudimentary verbalizer of _roles via a gpat (for lang or maodality specific orderings)
	 *
	 * @param {anne_neuron} $clumpNOBJ - the clump we want to print 
	 * @param {Object} $gpat - the language pattern we want to use
	 * @param {Array} $skipRoleNames - any roles we want to skip
     * @param {Boolean} $bDOSENTENCE - print in sentence mode
	 * @param {Boolean} $bDEBUG - degug the printing.
	 *
	 * @returns {anne_babytalk} - returns .output of sentence and .order of role tnids in the order they were printed from the gpat.
	 */
	 verbalizeInstActionClump=function($clumpNOBJ, $gpat=null, $skipRoleNames=[], $bDOSENTENCE=true, $bDEBUG=true){
        if ($gpat==null) $gpat = this.eng_pat1;
        var output = "";
		var parts = [];
		var order = [];
        if ($bDEBUG==true) console.log("%c print for clump"+$clumpNOBJ.text+" %cdosentence:"+$bDOSENTENCE, TALK.bggreenbig, TALK.bgblackyellow);
			
		var part_of_anne = null;	//EC type

		for (var p = 0; p < $gpat.length; p++) {
            var part = $gpat[p];
            var roles = part.roles;
			if ($bDEBUG==true) console.log("%c"+part.pat_part_name, TALK.bgblack);

			//loop the roles in the part roles array.
			//if the role is a string, it is just simple.
			//if it is an object, we have .p, .mod, .modfunct, and .role props.
            for (var r=0; r < roles.length; r++) {
                var role = roles[r];
				var pre = "";
				var mod = "";
				var modfunc = null;
				var det_func = null;
				var pre_default = "";
				if (typeof role !=="string") {
					pre = role.p;
					if (pre==undefined) pre="";
					mod = role.mod;
					if (mod==undefined) mod = null;
					modfunc = role.modfunc;
					if (modfunc==undefined) modfunc =null;
					det_func = role.det_func;
					if (det_func==undefined) det_func = null;
					pre_default = role.p_default;
					if (pre_default == undefined) pre_default = null;
					else {
						if (mod==null) pre = pre_default;	
					}
					role = role.role;	//set this last or we lose our others.
				}

				if ($skipRoleNames.length && $skipRoleNames.indexOf(role)!=-1) continue;

				var arr = $clumpNOBJ[role];
                if (ANNE.UTILS.isset(arr)==false) continue;

				var mod_arr = null;
				if (mod!=null && mod!="") {
					mod_arr = $clumpNOBJ[mod];
					if (ANNE.UTILS.isset(mod_arr)==false) {
						mod=null;
						mod_arr=null;
					}
				}

                var to = "";

				order = order.concat(arr);

				//if this is a role_action.. and the the tense is present.
				if (role=="role_action") {
					var actionNOBJ = MEMG.getNeuronByNid(arr[0].to_s);
					if ($clumpNOBJ.shift_lang==true) {
						//yet we have a role_tense shifter..  apply the tense.
						to = TALK.englishTensify(actionNOBJ.text, $clumpNOBJ.tense);
					}
					else to = TALK.verbalizeRolesArr(arr);

					part_of_anne = actionNOBJ.getECTop();
				}

				else if (modfunc!=null) {
					to = modfunc(arr, mod);
				}

				else if (mod_arr!=null) {
					var newmod = TALK.verbalizeRolesArr(mod_arr);
					to = newmod+" "+TALK.verbalizeRolesArr(arr);
				}

				else if (det_func!=null) {
					to = det_func(arr);
				}
		 
				else {
					to = TALK.verbalizeRolesArr(arr);
				}
				
				if (to=="") continue;	//do period still. whatever.
				if (pre!="" && pre!=undefined && pre!=null && pre.length>0) {
					parts.push({pre:pre, str:to, output:pre+" "+to, role:role, roles:arr});
					to = pre+" "+to;
				}
				else parts.push({pre:pre, str:to, output:to, role:role, roles:arr});
			
			
				if ($bDEBUG==true) console.log("%c"+role+" %cpre:"+pre+" %cto:"+to, TALK.bgblackyellow, TALK.bgred, TALK.bggreen);
			}
        }

		for (var p=0; p < parts.length; p++) {
			if (p == (parts.length-1)) output=output+parts[p].output+".";
			else output=output+parts[p].output+" ";
		}

		//make a sentence out of it?
		if ($bDOSENTENCE==true) {
			output = output.charAt(0).toUpperCase() + output.slice(1);
		}	
		if (ANNE.SETTINGS.debug()==true) console.log(output);
		$clumpNOBJ.eval = output;

		if ($bDEBUG==true) console.log("%c"+output, TALK.bggreen);

		return new anne_babytalk(output, order, part_of_anne);
    }


	/**
	 * 
	 * @param {*} $arr 
	 * @returns 
	 */
	verbalizeRolesArr = function($arr) {
		if ($arr==undefined || $arr==null || $arr.length==0) return "[ERROR]";
		if ($arr.length==1) {
			var rObj = $arr[0];
			if (rObj.nameless!=undefined) return "";
			if (rObj.role_text!=undefined) return rObj.role_text;
			rObj.role_text = MEMG.__MASTER_DISPLAY_RELON_TO(rObj);
			return rObj.role_text;
		}
		var str = "";
		var froms = [];
		var conj = "and";

		for (var r=0; r < $arr.length; r++) {
			var rObj = $arr[r];
			if (froms.length==0) froms.push(rObj.f);
			else if (froms.indexOf(rObj.f)!=-1) froms.push(rObj.f);
			else conj="or";
		}
		for (var r=0; r < $arr.length; r++) {
			var rObj = $arr[r];
			var role_text =  MEMG.__MASTER_DISPLAY_RELON_TO(rObj);
			str += role_text;
			rObj.role_text = role_text;
			if (r==$arr.length-2) str=str+" "+conj+" ";
			else if (r==$arr.length-1) ;
			else str=str+", ";
		}
		return str;
	}


	/**
	 * 
	 * @param {*} $arr 
	 * @returns 
	 */
	listifyStrArr = function($arr) {
		if (ANNE.UTILS.isset($arr)==false) return "";
		if ($arr.length==1) return $arr[0];
		if ($arr.length==2) return $arr[0]+" and "+$arr[1];
		var finalstr = $arr[0];
		for (var a=1; a <$arr.length; a++) {
			if (a==$arr.length-1) finalstr = finalstr+", "+$arr[a];
			else finalstr = finalstr+" and "+$arr[a];
		}
		return finalstr;
	}


	
	/** 
	 * print simple relon tags.
	 * 
	 * This TALK class is an utter disaster.. changed a gazillion times.
	 * Use this to do basic printing.  We have lots of rework to do just
	 * for this basic talk.  Would be better to get full PAS understanding.
	 */ 
	 simpleRelonTags = function ($rObj, $ttype=0, $bCOLORIZEBYTYPE=true, $showfromNid=false) {
		if ($rObj==null) {
			console.log("silly wabbit, you have to have a relon to print a relon");
			return;
		}
		if ($rObj.type==TYPE_AWARENESS) {
			var tt = TARGET_NID;
			if ($rObj.to_type==TO_TYPE_TXID) tt = TARGET_TXID;
			return this.verbalizeForAwareness($rObj, true);
		}

		var typeText = this.getTypeName($rObj.type, true, "rel-type");

		//willy wonka from name
		var fromNOBJ = null;
		var fromText = "";
		if ($rObj.from_nid!=null) {
			fromNOBJ = MEMG.getNeuronByNid($rObj.from_nid);	//all should be indexed
		}
		else if ($rObj.pending_from!=null) {
			fromNOBJ = MEMG.getNeuronByPendingNid($rObj.pending_from);
			if ($rObj.pending_from_debug!=undefined && $rObj.pending_from_debug.length>0) fromText = $rObj.pending_from_debug;//+"("+$rObj.pending_from+")";
			else fromText = $rObj.pending_from;
		}
		if (fromNOBJ==null) {
			console.log("NO FROM OBJ??");
			console.log($rObj);
			$rObj.error = true;
			return "[[error, bad relon]]";
		}
		else {
			if (fromNOBJ.emulated) {
				console.log("NO RELONS FROM EMULATED NEURONS -- they are not real!");
				return;
			}
			fromText = fromNOBJ.bestNameHTML();
		}
		var fromClass = this.getNeuronDisplayClass(fromNOBJ);

		//we want the name of the here even tho the from is actually
		//our nid (in the cmd relon)
		if ($rObj.debug_name!=undefined) {
			fromText = "(MAKE) "+$rObj.debug_name;
			fromClass = "neuron-ghost";
		}

		//reln text
		var relnNOBJ = MEMG.getNeuronByNid($rObj.reln);
		if (relnNOBJ==null) {
		    console.log("problem with the reln. It must exist in local mem");
			$rObj.error = true;
		    return "[[ERROR in RELN "+$rObj.reln+"]]"
		}
		var relonText = relnNOBJ.text;
		var relnClass = "";
		if (relnNOBJ.nid==MEM.hanai_in.nid || relnNOBJ.nid==MEM.hanai_in_all.nid) relnClass = "neuron-hanai";
		else if (relnNOBJ.nid==MEM.hanai_out.nid)  relnClass = "neuron-hanai";
		else relnClass = this.getNeuronDisplayClass(relnNOBJ);
		

		if ($showfromNid==true) {
			fromText+="("+fromNOBJ.nid+")";
		}

		//to type
		var toText = this.verbalizeRelonToByToType($rObj, $bCOLORIZEBYTYPE);
		var ttypemsg = "";
		if ($ttype!=-1) ttypemsg = TXB.printType($ttype);
		if ($ttype==ANNE.TXB.TXTYPE_MAKE_GLOBAL) {
			return "<span>&nbsp;"+typeText +" "+ttypemsg+" "+relonText+"&nbsp;<b>"+toText+"</b></span>";
		}
		else if ($ttype==ANNE.TXB.TXTYPE_MAKE_GLOBAL_LOCATION) {
			return "<span>&nbsp;"+typeText +" "+ttypemsg+" MAKE LOCATION <b>&nbsp;"+toText+"</b>&nbsp;in&nbsp;"+fromText+"</span>";
		}

		//to stuff
		var toClass = "rel-to";
		var toNOBJ = null;
		if ($rObj.to_type==TO_TYPE_ANNE) {
			if ($rObj.pending_to_s!=undefined) {
				var parts = $rObj.pending_to_s.split("_");
				if (parts.length>1) {
					toNOBJ = MEMG.getNeuronByNid(parts[1].trim());
					if (toNOBJ!=null) {
						toClass = this.getNeuronDisplayClass(toNOBJ);
					}
				}
			}
			else { 
				toNOBJ = MEMG.getNeuronByNid($rObj.to_s);
				if (toNOBJ!=null) {
					toClass = this.getNeuronDisplayClass(toNOBJ);
				}
			}
		}

		return "<span class='q-part "+fromClass+"' title='"+fromNOBJ.nid+"'>"+fromText+"</span>"+typeText+"<span class='q-part "+relnClass+"'>"+relonText+"</span><span class='q-part "+toClass+"'>"+toText+"</span><span class='q-part anne-tx-type'>*"+ttypemsg+"TX*</span>";
	}

	/**
	 * Helpful display to colorize things consistently thruout the app.
	 * This took a bit of time to arrive at -- for some time I tried to colorize
	 * class neurons red, instances blue, text green...
	 * but have now settled on ghost blue, anne yellow, text green, self red (maybe orange later), and hanai pink.
	 */
	getNeuronDisplayClass=function($nObj) {
		if ($nObj==null) {
			return "neuron-ghost";
		}
		
		if ($nObj.nid==MEM.USER._userNObj.nid) return "neuron-self";
		if ($nObj==undefined || $nObj==null) {
			return "neuron-ghost";
		}
		if ($nObj.ghost==true) return "neuron-ghost";
		else if (ANNE.isANNE($nObj.owned_by)==true) return "neuron-anne";
		else {
			if ($nObj.owned_by==ANNE.USER._userNObj.nid) {
				return "neuron-self";
			}
			else {
				return "neuron-other"  //black for someone else's neuron (not ANNE's).
			}
		}
	
	}


     makeSendQAlert=function($el, $alert_class, $ico, $msg, $id, $deleteCallback, $color="white") {
		alert("DO WE STILL DO THIS OR WAS THIS THE OLD QUEUE???")
        var alert = ANNE.A.ezAlert($el, $alert_class, "", $id);
		if ($ico!="") $msg = GUI_STUFF.getIco($ico) + $msg;

		ANNE.A.ezEl(alert, "div", "sendq-alert-text", $msg);
        var col2=ANNE.A.ezEl(alert, "div", "sendq-alert-actions");

        var innerspan=ANNE.A.ezEl(col2, "span", "sendq-alert-delete");
    
		var atagbtn = document.createElement("a");
		atagbtn.classList = "nav-link";
		atagbtn.setAttribute("role", "button");
		var delete_ico = GUI_STUFF.getIco('delete');
		atagbtn.innerHTML = delete_ico;
		atagbtn.href="#";
		
		if ($deleteCallback!=undefined) {
			atagbtn.addEventListener("click", e=>$deleteCallback(e));
		}

        innerspan.append(atagbtn);
		return alert;
    }


	/**
	 * 
	 * @param {*} $realTXID 
	 * @param {*} $relon 
	 * @param {*} $sep 
	 * @param {*} $bCOLORIZE 
	 * @returns 
	 */
	genShortRelonTextByTXID=function($realTXID, $relon=null, $sep="&nbsp;»&nbsp;", $bCOLORIZE=false) {
		var relon = $relon;
		if ($relon==null) relon = MEMG.lookupRelonByTXID($realTXID);
		if (relon==null) return "";
		var fNObj = MEMG.getNeuronByNid(relon.from_nid);
		var rNObj = MEMG.getNeuronByNid(relon.reln);
		var tNOBJ = null;
		if (relon.t) tNOBJ = MEMG.getNeuronByNid(relon.to_s);
		var to = "";
		if (tNOBJ!=null) to = tNOBJ.bestNameHTML();
		else if ($relon.to_type==TO_TYPE_TXID) to="{url}";
		else if ($relon.to_type==TO_TYPE_BLOB) to="{blob}";
		else to = relon.to_s;
		var from = fNObj.text;
		if (from=="") from = fNObj.nid;
		if ($bCOLORIZE==true) {
			return "<b>"+from+"</b>"+$sep+"<em class='text-red'>"+rNObj.text+"</em>"+$sep+"<b class='text-green'>"+to+"</b>";
		} 
		return "<b>"+from+"</b>"+$sep+"<em>"+rNObj.text+"</em>"+$sep+"<b>"+to+"</b>";
	}


	/**
	 * Uses the Q to print a friendly gui.
	 */
	printSanitizedAlertCmds=function($updateElID, $aQ, $bPRINTBALANCE=true) {
		var msgDiv = ANNE.A.setHTML($updateElID, ""); //qrb_txs_list_div
		
		if (ANNE.SETTINGS.debug()==true) console.log($aQ);
		
		//append bar to qrb_progress div
		var progDiv = document.createElement("div");
		progDiv.classList = "bg-light tx-proggy";
		progDiv.setAttribute("id", "tx_totals");
		ANNE.A.append('qrb_progress', progDiv);

		var progressBar = document.createElement("div");
		progressBar.setAttribute("id", "tx_progress_bar");
		progressBar.classList = "proggy-bar positive";
		progressBar.style.display = "inline-block!important";

		ANNE.A.append(progDiv, progressBar, true);

		for (var c=0; c < $aQ.length; c++) {
			var cmdObj = $aQ[c];
			var txtlbl = "";

			if (cmdObj.cmds_special.length && cmdObj.cmds_special.indexOf(TXB.CMD_POPULATE_WITH_VARS_ONLY_NO_SEND)!=-1) {
				continue;//no need to display this in the gui.
			}
			
			if (cmdObj.relon!=undefined && cmdObj.relon.warning!=undefined) {
				ANNE.A.ezAlert(msgDiv, "bg-txid-alert-q bg-warning", "tx", c+"x_warn",  cmdObj.relon.warning);
			}

			var msg = "";
			if (cmdObj.msg!="" && cmdObj.relon!=null && cmdObj.relon.debug_name==undefined) {
				msg = cmdObj.msg;
			}
			else if (cmdObj.relon!=undefined) {
				//$rObj, $ttype=0, $bCOLORIZEBYTYPE=true, $showfromNid=false
				msg = ANNE.TALK.simpleRelonTags(cmdObj.relon, cmdObj.TXTYPE, false);
			}

			
			if (cmdObj.relon!=undefined && cmdObj.relon!==null) {
				if (cmdObj.relon.type==TYPE_PARAM) {
					msg = msg +" <span class='tiny-text'>PARAM setter</span>";
				}
			}

			if (cmdObj.relon!=undefined && cmdObj.relon.sys_ref_txid!=undefined) {
				msg = msg +" <span class='tiny-text'>&nbsp;depends on "+cmdObj.relon.sys_ref_txid+"</span>";
			}

			if (cmdObj.cmds_special.indexOf(TXB.CMD_ENCRYPT_CHILD_TO_CMD)!=-1 
			|| cmdObj.cmds_special.indexOf(TXB.CMD_ENCRYPT_SELF_TO)!=-1) {
				msg = msg +"<span class='tiny-text'>&nbsp;ENCRYPTED AFTER NID IS ACQUIRED</span>";
			}

			var ico = "";
			ico = UTILS.formatNumber(cmdObj.totalAnnies/100)+ANNE.GUI_STUFF.getAnneImgHtml('16px');

			//friendly ico display in the q
			var ico3 = "";
			if (cmdObj.relon!=undefined) {
				ico3 = ANNE.TXB.getTxICOByType(cmdObj.TXTYPE, cmdObj.relon);
			}

			var txtlbl = "<span class='q-relon-link'>tx #"+(c+1)+"</span>"+ico3;
			var total = "<span class='q-flexright' style='text-align:right'>"+ico+"</span>";

			var finalmsg = txtlbl+msg+total;

			ANNE.A.ezAlert(msgDiv, "bg-txid-alert-q", "tx", c+"x",  finalmsg);
		}
	}


	/**
	*/
	printNids = function($nids, $title="") {
		if ($title!="") console.log("%c"+$title, ANNE.TALK.bgbigred);
		
		for (var n=0; n < $nids.length; n++) {
			var nObj = MEMG.getNeuronByNid($nids[n]);
			console.log("n="+n+" "+nObj.text);
		}	
	}

	
	/**
	 *
	 */
	updateSantizedAlertCmd=function($updateElID, $cmdObj, $current_index, $total, $badgeEl="") {
		var txid = $cmdObj.broadcast_txid;
		var percentFloat = ($current_index+1)/$total;
		var bFINAL = false;
		var remain = $total-($current_index+1);
		
		if (remain>0) ANNE.A.setHTML($badgeEl, remain);
		else ANNE.A.setHTML($badgeEl, "");
		
		if ($current_index+1==$total) bFINAL=true;

		var percent = Math.ceil(percentFloat*100);
		ANNE.A.setStyle("tx_progress_bar", {width:percent+"%"}, true);
		var msgDiv = document.getElementById($updateElID);

		//try to track with the element getting updated.
		
		if (($cmdObj.skipped==undefined || $cmdObj.skipped==false) && txid!=undefined && txid.length>0) {
			var msg = "";
			if ($cmdObj.msg=="") msg = "success! ";
			else msg = $cmdObj.msg;
		
			var el = document.getElementById('tx_'+$current_index+'x');
			if (el==null) return;

			msgDiv.scrollTop = el.offsetTop-400; //msgDiv.scrollHeight;
		
			var tx_alert = ANNE.A.addRemoveClass(el, "bg-green-themeDARK", "bg-txid-alert-q");
			
			//update the txid #n with the txid ico.
			var txidcol = tx_alert.querySelector(".q-relon-link");
			if (txidcol!=null && txid!="") {
				ANNE.MEM_UDF.makeRelonViewLink(txidcol, txid, true, 12, true);
			}

			//if this was the last $cmdObj, print the complete transactions.  :)
			if (bFINAL==true) ANNE.A.ezAlert(msgDiv, "bg-green-themeDARK", "", "",  "<b>"+$total+"</b>&nbsp;Transactions Complete");
			return;
		}

		else if ($cmdObj.skipped!=undefined && $cmdObj.skipped==true) {
			var msg = "skipped::";
			if ($cmdObj.skipped_reason!=undefined) msg += $cmdObj.skipped_reason;
			var el = document.getElementById('tx_'+$current_index+'x');
			if (el==null) return;

			msgDiv.scrollTop = el.offsetTop-400; //msgDiv.scrollHeight;
		
			var tx_alert = ANNE.A.addRemoveClass(el, "bg-red", "bg-txid-alert-q");
		}
	}



	/*** printRelonByID ************************************** xx/xx/202x ***
	******************************************************** Mr.Scatman ****/ 
	printRelonByID=function($id) {
		var rObj = MEMG.lookupRelonByTXID($id);
		if (rObj) this.printRelon(rObj);
	}

	//here we just want the names.
	printNeurons = function($nObjs) {
		for (var n=0; n < $nObjs.length; n++) {
			console.log($nObjs[n].bestNameHTML());
		}
	}

	printNObjsList = function($nObjs) {
		for (var n=0; n < $nObjs.length; n++) { console.log($nObjs[n].bestName());}
	}

	/*** printNObjs ****************************************** xx/xx/202x ***
	******************************************************** Mr.Scatman ****/
	printNObjs = function ($nObjs) {
		for (var n=0; n < $nObjs.length; n++) {
			console.log("***********************************");
			var nObj = $nObjs[n];
			console.log("for neuron: "+nObj.nid);
			if (nObj.mine==true) console.log("mine=true!");
			else console.log("not mine");

			if (nObj.temp==true) console.log("temp neuron");
			if (nObj.ghost==true) console.log("local ghost neuron");
			if (nObj.preloaded==true) console.log("preloaded");
			if (nObj.emulated==true) console.log("emulated neuron");
			if (nObj.owned_by!=undefined) console.log("owned by "+nObj.owned_by);
			if (nObj._gt!=undefined) {
				console.log("gt "+MEMG.getNeuronByTempNid(nObj._gt.t).text);
				this.printRelon(nObj._gt);
			}
			this.printRelons(nObj.outs);
		}
	}


	printFromTempNidArray=function($tnids) {
		for (var t=0; t < $tnids.length; t++) {
			var nObj = MEMG.getNeuronByTempNid($tnids[t]);
			if (nObj==null) continue;
			console.log("for neuron: "+nObj.nid);
			if (nObj.temp==true) console.log("temp neuron");
			if (nObj.ghost==true) console.log("local ghost neuron");
			if (nObj.preloaded==true) console.log("preloaded");
			if (nObj.emulated==true) console.log("emulated neuron");
			if (nObj.owned_by!=undefined) console.log("owned by "+nObj.owned_by);
			if (nObj._gt!=undefined) {
				console.log("gt "+MEMG.getNeuronByTempNid(nObj._gt.t).text);
				this.printRelon(nObj._gt);
			}
			this.printRelons(nObj.outs);
		}
	}


	printOnlyRelons = function($TYPE, $aRelons, $plainText=true, $bDOTEMPNIDS=false, $consoleColor=-1) {
		for (var r = 0; r < $aRelons.length; r++) {
			if ($aRelons[r].type!=$TYPE) continue;
			this.printRelon($aRelons[r], $plainText, $bDOTEMPNIDS, $consoleColor);
		}
	}

	/**
	 * Print an array of relons.
	 * 
	 * @param {Array} $aRelons 
	 * @param {Boolean} $plainText 
	 * @param {Boolean} $bDOTEMPNIDS 
	 * @returns 
	 */
	printRelons = function ($aRelons, $plainText=true, $bDOTEMPNIDS=false, $consoleColor=-1) {
		if (ANNE.UTILS.isset($aRelons)==false) {
			console.log("no relons to print");
			return;
		} 

		console.log($aRelons);

		for (var r = 0; r < $aRelons.length; r++) {
			this.printRelon($aRelons[r], $plainText, $bDOTEMPNIDS, $consoleColor);
		}
		return;
	}


	printTNIDS = function($tnids){
		var texts=[];
		for (var t=0; t < $tnids.length; t++) {
			var nOBJ = MEMG.getNeuronByTempNid($tnids[t]);
			console.log(nOBJ.text);
			texts.push(nOBJ.text);
		}
		console.log(texts.join(","));
		return texts.join(",");
	}



	/*** selfRelon ******************************************* xx/xx/2020 ***
	******************************************************** Mr.Scatman ****/ 
	selfRelon = function ($relon, $bYOU=true) {
		var relnNOBJ = MEMG.getNeuronByTempNid($relon.r);
		var toObj = MEMG.getNeuronByTempNid($relon.t);
		var totxt="";
		if (toObj!=null) totxt = toObj.text;
		else totxt = $relon.to_s;

		var your = "Your";
		if ($bYOU==false) your = "Their";
		var you = "You";
		if ($bYOU==false) you = "They";

		var msg = "<span class='text-primary'> "+you+"</span>";
		if ($relon.type==TYPE_AWARENESS) {
			
			var toTxt = "";
			if ($relon.to_type==TO_TYPE_ANNE) {
				toTxt = this.verbalizeForAwareness($relon);
			}
			else toTxt = $relon.to_s;
			return your + " "+relnTxt + " "+toTxt;
		}

		else if ($relon.type==TYPE_BE) {
			if ($relon.to_type==TO_TYPE_STRING) {
				var your = "<span class='text-primary'> "+your+"</span>";
				if ($relon.r==MEM.name.temp_nid) return your+" <span class='text-danger'>name </span>is <span class='text-success-lbl'>"+totxt+"</span>";
				else if ($relon.r==MEM.aka.temp_nid) return msg+" are <span class='text-danger'>also known as </span>is <span class='text-success-lbl'>"+totxt+"</span>";
			}
			msg = msg +" are ";
			var determiner = this.getDet(totxt);
			return msg +determiner+" <span class='text-danger'>"+totxt+"</span>";
		}
		else if ($relon.type==TYPE_SYS) {
			if (toObj!=null) {
				if ($relon.r==MEM.necrosis.temp_nid) {
					"<span class='text-danger-lbl'>"+relnNOBJ.text+" (delete) </span> <span class='text-success'>"+totxt+"</span>";;
				}
				else return relnNOBJ.text+" = "+totxt;
			}
			else return relnNOBJ.text+" = "+$relon.to_s; 
		}
		else if ($relon.type==TYPE_HAS) {
			msg = msg+" have ";	
			var determiner = this.getDet(relnNOBJ.text);
			var reln_txt = "<span class='text-danger'>"+relnNOBJ.text+"</span>";
			msg = msg+determiner+" "+reln_txt+" of ";

			var toStr = "";
			if (toObj!=null) toStr = totxt;
			else toStr = $relon.to_s;

			var to_txt = ""
			if (toObj!=null && toObj.bTERMINATING_TYPE==true) to_txt = "<span class='text-success'>"+toStr+"</span>";
			else to_txt = "<span class='text-danger'>"+toStr+"</span>";
			return msg + to_txt;
		}
	}


	printListByReln = function($outs, $r, $excludeTypes=[TYPE_PARAM], $sep=", ") {
		var relons = MEMG.getRelonsByTempR($outs, $r, $excludeTypes);
		if (relons.length==0) return "";
		return MEMG.getTextListByRelonTos(relons, $sep);
	}


	colorizeByPol=function($nObj) {
		if ($nObj.polarity==undefined) MEM.polarity.set_value($nObj);
		var fclass = "";
		if ($nObj.polarity==1) fclass = "text-success-lbl";
		else if ($nObj.polarity==-1) fclass = "text-danger-lbl";
		else if ($nObj.polarity==0) fclass = "text-warning-lbl";
		return "<span class='q-part "+fclass+"'>"+$nObj.text+"</span>";
	}


	printEzName = function($nObj) {
		var text = $nObj.text;
		var be = $nObj.be;
		if (text=="") {
			return "<em>(unnamed) "+be+"</em> <span class='raw-address'>"+$nObj.nid+"</span>";
		}

		return "<em>"+be+"</em> <span style='color:black; font-weight:bold'>"+text+"</span>&nbsp;(<span class='raw-address'>"+$nObj.nid+"</span>)";
	}



	/**
	 * Get a determiner for a given string.
	 * 
	 * Have to go to phonemes to get accurate with this.
	 * Not doing that now.  going the cheap and inaccurate road.
	 * 
	 * @author {Mr.Scatman}
	 */ 
	getDet=function ($str) {
		var str =$str.toLowerCase();
		var char = str.substring(0,1);
		var aeiou = "aeiou";
		if (aeiou.indexOf(char)!=-1) return "an";
		return "a";	
	}

	/**
	 * Another rudimentary det function in lieu of using NLP in this app.
	 * (that comes later with a more robust vocab and understanding of lang in the circuits..
	 * ...tho.. technically... we can use the NLP classes to parse now)
	 * @param {*} $nObj 
	 * @returns 
	 */
	getDet2=function ($nObj) {
		var str = $nObj.text.toLowerCase();

		if ($nObj.ntype==NTYPE_INST) {
			if ($nObj.bes.indexOf(MEM.person.temp_nid)!=-1 || $nObj.bes.indexOf(MEM.location.temp_nid)!=-1) return $nObj.text;
			return "the "+$nObj.text;
		}

		//an elephant, an apple, a peach, a car, a dog.
		else {
			var char = str.substring(0,1);
			var aeiou = "aeiou";
			if (aeiou.indexOf(char)!=-1) return "an "+$nObj.text;
			return "a "+$nObj.text;		
		}
	
	}
	


	/** for later.. */
	toSoundEncoding(word) {
		return word.toLowerCase().split('')
		.map(function(c) { return encodeCharacter(c);})
		.join('')
		.replace('ng', 'n')
		.replace('nn', 'n')
		.replace('oo', 'o')
		.replace('oo', 'o')
		.replace('tt', 't')
		.replace('ol', 'l')
		.slice(0,4);
	}
	

	/** for later.. */ 
	soundsLike=function($w1, $w2) {
		return this.toSoundEncoding($w2) === this.toSoundEncoding($w2);
	}


	/** for later.. */ 
	toSoundEncoding=function($w) {
		return $w.toLowerCase().split('')
		.map(function(c) {return this.encodeCharacter(c);})
		.join('')
		.replace('ng', 'n')
		.replace('nn', 'n')
		.replace('oo', 'o')
		.replace('oo', 'o')
		.replace('tt', 't')
		.replace('ol', 'l')
		.slice(0,4);
	}
	

	/** for later.. */
	encodeCharacter=function(c) {
		if ('kgq'.indexOf(c) >= 0) return 'k'; // velar stops
		if ('bpfv'.indexOf(c) >= 0) return 'p'; // labials
		if('td'.indexOf(c) >= 0) return 't'; // alveolars
		if ('szjc'.indexOf(c) >= 0) return 's'; // shushes
		if ('mn'.indexOf(c) >= 0) return 'm'; // nasals
		if ('lr'.indexOf(c) >= 0) return 'l'; // glides
		if ('aeiou'.indexOf(c) >= 0) return 'o'; // vowels
		if ('yhw'.indexOf(c) >= 0) return '';
		return '';
	}
	

	/**
	 * Quick n dirty ending printer for frenly day printing of dates.
	 */
	lazyDayIntEnding = function($int) {
		if ($int==1 || $int==21 || $int==31) return $int+"st";
		else if ($int==2 || $int==22) return $int+"nd";
		else if ($int==3 || $int==23) return $int+"rd";
		else return $int+"th";
	}


	/**
	 * poor man's printing -- no intelligence involved.	
	 * quite rudimentary -- but for now, it works.
	 */
	humanReadableRelon = function ($relon, $bNOFONT=false) {
		//redo this.. for now, just print relon.
		return this.printRelon($relon, true, false);
	}


	/**
	 */
	whoOwnsPhrase = function($who, $nObj=null) {
		if ($who=="anne") return "A.N.N.E. owns this neuron";
		else if ($who=="app") return "This app owns this neuron";
		else if ($who=="me") {
			if ($nObj.temp_nid==USER._userNObj.temp_nid) return "<span class='text-primary'>This IS you!</span>";
			else return "<span class='text-primary'>You owns this neuron!</span>";
		}
		else if ($who=="worldview") return "This is in your <span class='text-primary'>worldview</span>";
		
		return "**This neuron is not in your worldview**";
	}


	
	/**
	 */ 
	 verbalizeForAwareness = function ($relon, $bTAGIFY=true) {
		if ($relon.to_type!=TO_TYPE_ANNE) return " [[[don't use this w/ non neuron tos]]] ";
		var toNOBJ = MEMG.getNeuronByTempNid($relon.t);
		var rNOBJ = MEMG.getNeuronByTempNid($relon.r);
		var relnTxt = this.colorizeByPol(rNOBJ);

		//opinionz are just keywords..
		//if (rNOBJ.bes[0]==MEM.symbol.temp_nid) {
			//var determiner = this.getDet(relnTxt);
			//if ($bTAGIFY==false) return "I think "+toNOBJ.bestNameHTML()+" is "+determiner+" "+relnTxt;	
			//return "<span class='type-aware'>AWARE</span><span class='q-part rel-from'>I</span> <span class='q-part'>think </span><span class='q-part'>"+toNOBJ.bestNameHTML()+"</span><span class='q-part'>is</span> "+relnTxt;	
		//}

		//else 
		if (rNOBJ.nid==MEM.rememberz.nid) {
			if ($bTAGIFY==false) return "rememberz "+toNOBJ.bestNameHTML();
			return "<span class='type-aware'>AWARE</span><span class='q-part rel-reln'>rememberz</span><span class='q-part reln-to'>"+toNOBJ.bestNameHTML()+"</span>";
		}

		//the rest are pos/neg/neu feelings (buckets) or feelz (most refined expression of feeling)
		else {
			if ($bTAGIFY==false) return ""+toNOBJ.bestNameHTML()+" makes me FEELZ "+relnTxt;
			return "<span class='type-aware'>AWARE</span><span class='q-part rel-to'>"+toNOBJ.bestNameHTML()+"</span><span class='q-part'>makes</span><span class='q-part rel-from'>me</span>"+relnTxt;
		}
	}


	/**
	 */
	 printAwarenessForDisplay = function($relnNOBJ, $tnid) {
		var tnid = parseInt($tnid);
		var targetNOBJ = MEMG.getNeuronByTempNid(tnid);
		if (targetNOBJ==null) return "fix this";
		if (targetNOBJ.bes[0]==MEM.opinionz.temp_nid) {
			var determiner = this.getDet($relnNOBJ.text);
			return "I think "+toNOBJ.bestNameHTML()+" is "+determiner+" "+$relnNOBJ.text;	
		}
		else if ($relnNOBJ.nid==MEM.rememberz.nid) return "rememberz "+targetNOBJ.bestNameHTML();	
		return targetNOBJ.bestNameHTML()+" makes me FEELZ "+$relnNOBJ.bestNameHTML();
	}


	//tells us what something is from a text search.
	//this function is used as a priority when searching for things
	//we prioritize YOU first, then ANNE, then temp-worldview-pass (review this), then outside the network.
	verbalizePriorityFromTallySummary=function($tally, $textVal, $parentNObj=null, $prop="name") {
		var who = ""
		var has = "has";
		if ($tally.me) {
			who = "YOU";
			has = "have";
		}
		else if ($tally.anne) who = "A.N.N.E.";
		else if ($tally.app) who = "THIS '"+MEM_UDF.THIS_APP.text.toUpperCase()+"' APP";
		else if ($tally.worldview) who = "An adopted neuron in your network";
		else if ($tally.wv_pass) who = "A temporary-worldview-pass neuron"; 
		else if ($tally.other) who = "Someone outside of your network";
		who = "<span style='color:white'>"+who+"</span>"+"&nbsp;"+has;
		var search_msg = who+"&nbsp;already created a neuron with the same '"+$prop+"' <span class='text-success'>&nbsp;'"+$textVal+"'</span>";
		if ($parentNObj != null) search_msg=search_msg+"&nbsp;with parent class <span class='text-success'>&nbsp;'"+$parentNObj.text+"'</span>";
		msg = "As a friendly reminder;&nbsp;"+search_msg+"!  Are you sure you want to create another one?"; 
		return msg;
	}


	/**
	 */
	 typeTalk=function($type, $bTEXTONLY=false, $invertBG=false) {
		if ($bTEXTONLY) return this.type_pre[parseInt($type)];
		var invert = "";
		if ($invertBG==true) invert="-invert";
        var type_string="";
        if ($type==TYPE_SYS_MSG) type_string="<span class='type-sys"+invert+"'>sys msg</span>";
        else if ($type==TYPE_SYS) type_string="<span class='type-sys"+invert+"'>sys</span>";
        else if ($type==TYPE_BE) type_string="<span class='type-be"+invert+"'>be</span>";
        else if ($type==TYPE_PARAM) type_string="<span class='type-param"+invert+"'>param</span>";
        else if ($type==TYPE_HAS) type_string="<span class='type-has"+invert+"'>has</span>";
        else if ($type==TYPE_AWARENESS) type_string="<span class='type-aware"+invert+"'>aware</span>";
        else if ($type==TYPE_CLUMP) type_string="<span class='type-clump"+invert+"'>clump</span>";
		else if ($type==TYPE_ROLE) type_string="<span class='type-role"+invert+"'>role</span>";
		else if ($type==TYPE_BRAIN) type_string="<span class='type-brain"+invert+"'>brain</span>";
		else if ($type==TYPE_EPISODIC) type_string="<span class='type-episodic"+invert+"'>episodic</span>";
		else if ($type==TYPE_EXPERIENCE) type_string="<span class='type-experience"+invert+"'>experience</span>";
		else if ($type==TYPE_INTERNAL_PAY) type_string="<span class='type-pay"+invert+"'>internal payment</span>";
		else if ($type==TYPE_BROADCAST_PAY) type_string="<span class='type-pay"+invert+"'>broadcast payment</span>";
		else if ($type==TYPE_BRAIN120) type_string="<span class='type-brain120"+invert+"'>brain120 backlink</span>";
		else type_string = "unhandled ("+$type+")";
		return type_string;
    }


	/**
	 * 
	 * @param {*} $type 
	 * @param {*} $fromNOBJ 
	 * @param {*} $relnNOBJ 
	 * @param {*} $toType 
	 * @param {*} $toText 
	 */
	quickRelonLinePrint=function($type=null, $fromNOBJ=null, $relnNOBJ=null, $toType=null, $toText=null) {
		var parts = [];
		if ($type!=null) parts.push("["+ANNE.TALK.typeTalk($type)+"]");
		if ($fromNOBJ!=null) parts.push($fromNOBJ.bestNameHTML()+" ->");
		if ($relnNOBJ!=null) parts.push($relnNOBJ.bestNameHTML());
		if ($toType!=null) parts.push("["+TALK.toTypeTalk($toType)+"] ->");
		if ($toText!=null) parts.push($toText);
        return parts.join(" ");
	}

	
	/**
	 */
    toTypeTalk=function($to_type, $bI=false) {
        var to_type_string="";
        if($to_type==TO_TYPE_SEED) to_type_string="encrypted seed";
        else if($to_type==TO_TYPE_ANNE) to_type_string="address";
        else if($to_type==TO_TYPE_TXID) to_type_string="txid";
        else if($to_type==TO_TYPE_URL) to_type_string="url";
        else if($to_type==TO_TYPE_NUMBER) to_type_string="number";
        else if($to_type==TO_TYPE_DATE) to_type_string="date";
        else if($to_type==TO_TYPE_DATETIME) to_type_string="datetime";
        else if($to_type==TO_TYPE_STRING) to_type_string="string";
        else if($to_type==TO_TYPE_BLOB) to_type_string="blob";
        else if($to_type==TO_TYPE_TIME) to_type_string="time";
        else if($to_type==TO_TYPE_BOOLEAN) to_type_string="boolean";
        else if($to_type==TO_TYPE_ENCR_ANNE) to_type_string="encr addr";
        else if($to_type==TO_TYPE_ENCR_TXID) to_type_string="encr txid";
        else if($to_type==TO_TYPE_ENCR_URL) to_type_string="encr url";
        else if($to_type==TO_TYPE_ENCR_NUMBER) to_type_string="encr #";
        else if($to_type==TO_TYPE_ENCR_DATE) to_type_string="encr date";
        else if($to_type==TO_TYPE_ENCR_DATETIME) to_type_string="encr dtime";
        else if($to_type==TO_TYPE_ENCR_STRING) to_type_string="encr str";
        else if($to_type==TO_TYPE_ENCR_BLOB) to_type_string="encr blob";
        else if($to_type==TO_TYPE_ENCR_TIME) to_type_string="encr time";
        else if($to_type==TO_TYPE_ENCR_BOOLEAN) to_type_string="encr bool";
        

        if($bI==true) return "<i>"+to_type_string+"</i>";
        else return to_type_string;
    }


	/**
	 */
	use_d_pat=function($d_pat, $nObj, $bDOLINKS=true, $linkType=LINKS_NEURON_ANNET) {
		if ($d_pat==undefined || $d_pat==null || $d_pat.length==0) return;
		var msg = "";
		for (var d=0; d < $d_pat.length; d++) {
			var pat = $d_pat[d];
			var prop = pat.prop;
			var pre = pat.pre;
			var post = pat.post;
			var bALT = false;
			if (prop==undefined) continue;
			var val = $nObj[prop];
			if (val==undefined || val==null) {
				if (pat.alt_lbl!=undefined) {
					val = pat.alt_lbl;
					bALT = true;
				}
				else continue;
			}
			if ($bDOLINKS==true && bALT==false) {
				var valNOBJ = MEMG.getNeuronByTempNid(val);
				if (valNOBJ==null) {
					if (pat.cancel_prev!=undefined) return "";
					continue;
				}
				val = LINKS.makeNeuronLink(valNOBJ, $linkType);
				if (pat.show_aid!=undefined && pat.show_aid==true) {
					val = val +"&nbsp;("+valNOBJ.nid+")";
				}
			}

			if (pre!="") {
				if (msg!="") msg = msg+"&nbsp;"+pre;
				else msg = pre;
			}
			if (msg=="") msg = val;
			else msg = msg+"&nbsp;"+val;
			
			if (post!="") msg = msg+"&nbsp;"+post;
		}

		return msg.trim();
	}

	printTempNids = function($tnids) {
		for (var t=0; t < $tnids.length; t++) {
			var nObj = MEMG.getNeuronByTempNid($tnids[t]);
			if (nObj==null) continue;
			console.log(nObj.text, nObj);
		}
	}

	/**
	 * plaintext no longer does anything.
	 * this function is a DISASTER, fay.  -dlm
	 */
	printRelon = function ($rObj, $plainText=false, $bDOTEMPNIDS=false, $consoleColor=-1) {
		if ($rObj==null) {
			console.log("silly wabbit, you have to have a relon to print a relon");
			return;
		}

		//console.log($rObj);
		var typeText = this.getTypeName($rObj.type);
		var fromNOBJ = MEMG.getNeuronByNid($rObj.from_nid);	//all should be indexed
		if (fromNOBJ==null) {
			return;
		}

		var emulated = "";
		var fromText = fromNOBJ.text;
		if (fromText=="") fromText = fromNOBJ.nid;
		var relnNOBJ = MEMG.getNeuronByNid($rObj.reln);
		var relonText = relnNOBJ.text;
		if ( fromNOBJ.emulated) emulated = "(emulated) ";
		var relonStr = emulated+typeText + " from:" + fromText+ "  reln:"+relonText;
		var toText = this.verbalizeRelonToByToType($rObj);
		relonStr += " to:"+toText;

		var toSPrintOnly = this.toSPrint($rObj);

		var freeze = fromNOBJ.freeze;
		if (freeze==1) console.log("%cfrom nid froze", TALK.bgivoryblue);

		var ghost = "";
		if ($rObj.ghost==true) ghost="(gh)";

		var sysreftxid=""
		if ($rObj.sys_ref_txid!=undefined) sysreftxid=" sysref:"+$rObj.sys_ref_txid;

		var annies=""
		if ($rObj.annies!=undefined) annies=" annies:"+$rObj.annies;
		
		var fhi="";
		if ($rObj.flathexint!=undefined) fhi=" fhi:"+$rObj.flathexint;
		
		if ($consoleColor!=-1) {
			console.log("%cg"+ghost+"txid:"+$rObj.txid+"["+typeText+"]"+"["+fromText+"]"+"["+relonText+"]"+"["+toSPrintOnly+"] "+sysreftxid+annies+fhi, $consoleColor);
		}
		else {
			console.log(ghost+"txid:"+$rObj.txid+"["+typeText+"]"+"["+fromText+"]"+"["+relonText+"]"+"["+toSPrintOnly+"] "+sysreftxid+annies+fhi);
		}

		if ($plainText==true) return emulated+"[" + typeText+"]"+fromText+relonText+toSPrintOnly;
		return relonStr;
	}


	toSPrint=function($rObj){
		if ($rObj.pending_to_s!=undefined && $rObj.to_s==null) {
			if ($rObj.pending_to_debug!=undefined) return "*pending* "+$rObj.pending_to_debug;
			else return "pending nid/txid";
		}
		var toText = this.getToTextByToType($rObj.to_type, $rObj.to_s, false);
		return toText;
	}


	/**
	 */
	verbalizeRelonToByToType = function ($rObj, $bCOLORIZEBYTYPE=true) {
		if ($rObj.pending_to_s!=undefined && $rObj.to_s==null) {
			if ($rObj.pending_to_debug!=undefined) return "*pending* "+$rObj.pending_to_debug;
			else return "pending nid/txid";
		}
		var toText = this.getToTextByToType($rObj.to_type, $rObj.to_s, $bCOLORIZEBYTYPE);
		return toText;
	}


	/**
	 */ 
	getToTextByToType = function ($to_type, $to_s, $colorize=true) {
		var text = $to_s;
		if ($to_type == TO_TYPE_ANNE) {
			var nObj = MEMG.getNeuronByNid($to_s);
			if (nObj) {
				if ($colorize==true) text = nObj.bestNameHTML();
				else text = nObj.bestName();
			}	
		}
		if ($colorize) return "<span class='to-type tt-"+$to_type+"'>"+text+"</span>";
		else return text;
	}


	/**
	 */
	getTypeName = function ($type, $bDOCLASSHTML=false, $addClass="") {
		if ($bDOCLASSHTML==false) {
			if ($type == TYPE_SYS || $type=="sys") return 'SYS';
			if ($type == TYPE_BE || $type=="be") return 'BE';
			if ($type == TYPE_PARAM || $type=="para") return 'PARAM';
			if ($type == TYPE_HAS || $type=="has") return 'HAS';
			if ($type == TYPE_AWARENESS || $type=="awareness") return 'AWARENESS';
			if ($type == TYPE_SYS_MSG || $type=="msg") return  'SYS MSG';
			if ($type == TYPE_EXPERIENCE || $type=="exper") return  'EXPER';
			if ($type == TYPE_CLUMP || $type=="clump") return  'CLUMP';
			if ($type == TYPE_ROLE || $type=="role") return  'ROLE';
			if ($type == TYPE_BRAIN || $type=="brain") return  'BRAIN';
			if ($type == TYPE_INTERNAL_PAY || $type=="internal payment") return  'INTERNAL PAYMENT';
		}
		else {
			if ($type == TYPE_SYS || $type=="sys") return '<span class="type-sys '+$addClass+'">SYS</span>';
			if ($type == TYPE_BE || $type=="be") return '<span class="type-be '+$addClass+'">BE</span>';
			if ($type == TYPE_PARAM || $type=="param") return '<span class="type-param '+$addClass+'">PARAM</span>';
			if ($type == TYPE_HAS || $type=="has") return '<span class="type-has '+$addClass+'">HAS</span>';
			if ($type == TYPE_AWARENESS || $type=="awareness") return '<span class="type-awareness '+$addClass+'">AWARE</span>';
			if ($type == TYPE_SYS_MSG || $type=="msg") return '<span class="type-sys-msg '+$addClass+'">SYS MSG</span>';
			if ($type == TYPE_EXPERIENCE || $type=="exper") return '<span class="type-experience '+$addClass+'">EXPER</span>';
			if ($type == TYPE_CLUMP || $type=="clump") return '<span class="type-clump '+$addClass+'">CLUMP</span>';
			if ($type == TYPE_ROLE || $type=="role") return '<span class="type-role '+$addClass+'">ROLE</span>';
			if ($type == TYPE_BRAIN || $type=="brain") return '<span class="type-brain '+$addClass+'">BRAIN</span>';
			if ($type == TYPE_INTERNAL_PAY || $type=="internal payment") return '<span class="type-brain '+$addClass+'">INTERNAL PAYMENT</span>';
		}

		//anything 100 and over is typebrain 0, 1, 2 etc
		if ($type>=20 && $type<120) return "BRAIN"+(parseInt($type));
		return 'unknown '+$type;
	}


	/**
	 */
	pluralizeNounIf = function($mixed, $baseWord, $plural, $singularPostFix="") {
		if ((Array.isArray($mixed)==true) && ($mixed.length>1 || $mixed.length==0)) return $baseWord+$plural;
		else if (typeof $mixed==="number" && ($mixed>1 || $mixed==0)) return $baseWord+$plural;
		return $baseWord+$singularPostFix;
	}

	pluralizeNounIf2 = function($mixed, $baseWord, $pluralwd, $singularPostFix="") {
		if ((Array.isArray($mixed)==true) && ($mixed.length>1 || $mixed.length==0)) return $pluralwd;
		else if (typeof $mixed==="number" && ($mixed>1 || $mixed==0)) return $pluralwd;
		return $baseWord+$singularPostFix;
	}

	/**
	 */
	pluralizeNP = function ($arr, $baseWord, $plural, $conj="or", $nestedProp="content") {
		if ($arr.length>1) return $baseWord+$plural+" "+this.listify($arr, $conj, $nestedProp);
		else  {
			if ($nestedProp!="") return $baseWord+"&nbsp;<span class='text-success-lbl'>"+$arr[0][$nestedProp].text+"</span>";
			else return $baseWord+"&nbsp;<span class='text-success-lbl'>"+$arr[0].text+"</span>";
		}
	}


	/**
	 */
	listify=function($arr, $conj, $nestedProp="content"){
		var list = "";
		for (var a=0; a < $arr.length; a++) {
			var sep = "";
			if ((a+1)==$arr.length) sep = "</span>&nbsp"+$conj+"&nbsp;<span class='text-success-lbl'>";
			else if (a>=1) sep = ",&nbsp;";
			else list = "<span class='text-success-lbl'>";
			
			if ($nestedProp) list += sep+""+$arr[a][$nestedProp].text+"";
			else list += sep+"<b>"+$arr[a].text+"</b>";
		}
		return list+"</span>";
	}


	/**
	 */
	prounounPick=function($arr, $pr, $prpl) {
		if ($arr.length>1) return $prpl;
		else return $pr;
	}


	//To make regular nouns plural, add ‑s to the end. ...
	//If the singular noun ends in ‑s, -ss, -sh, -ch, -x, or -z, add ‑es to the end to make it plural. ...
	//In some cases, singular nouns ending in -s or -z, require that you double the -s or -z prior to adding the -es for pluralization.
	pluralizeNoun=function($text) {
		var rev = $text.split("").reverse().join("");
		var c0 = rev[0];
		var c1 = "";
		if (rev.length>1) c1 = rev[1];
		if (c0=="s" || c0=="x" || c0=="z") {
			if (c0=="s" && c1=="s") return $text+"s";
			if (c0=="s") return $text+"ses";
			if (c0=="z") return $text+"zes";
			return $text+"es";
		}
		if (c0=="h" && (c1=="s" || c1=="c")) return $text+"es";
		return $text + "s";
	}


	/**
	 */
	beState = function($array) {
		if ($array.length==0) return "are";
		else if ($array.length==1) return "is";
		return "are";
	}

	/**
	 * make an english list phrase of "x, y, z and a" from an array of [x, y, z, a]
	 */
	listifyArray=function($arr) {
		if ($arr.length==0) return "";
		if ($arr.length==1) return $arr[0];
		var str = $arr[0];
		for (var a=1; a < $arr.length-1; a++) {str = str+", "+$arr[a];}
		str = str+ " and "+$arr[($arr.length-1)];
		return str;
	}


	/**
	 * Just some helpful dumb msgs popped in here for certain relons.
	 * We can display these in the relons page to inform the user a bit more
	 * about what this relons is/was for.
	 */
	basicExplain = async($rObj)=> {
		var rel_sponsor_msg="";

		if ($rObj.relon_sponsor!=undefined && $rObj.relon_sponsor!=null && $rObj.relon_sponsor!="") {
			rel_sponsor_msg = "  This was originally a collab";
		}

		if ($rObj.from_nid==MEM.___SELF___ && $rObj.reln==MEM.BLOCK_WINNER.nid) {
			return "This is a block reward winner.";
		}
		if ($rObj.type==TYPE_BROADCAST_PAY && $rObj.reln==MEM.annies.nid && $rObj.to_type==TO_TYPE_ANNE) {
			return "This was a broadcast payment relon.";
		}
		if ($rObj.ghost==true && $rObj.to_type>=TO_TYPE_ENCR_ANNE) {
			var fromNOBJ = MEMG.getNeuronByTempNid($rObj.f);
			if (fromNOBJ.ghost==true) return "This to value will be encrypted at run-time when broadcast.";
		}
		if ($rObj.r==MEM.anj_accepted.temp_nid) return "This is an accepted anjucator invite.";
		if ($rObj.r==MEM.anj_rejected.temp_nid) return "This is an declined anjucator invite.";
		if ($rObj.reln==MEM.anjucator.nid) {
			if ($rObj.sys_ref_txid!=undefined && $rObj.sys_ref_txid.length) {
				if ($rObj.to_type==TO_TYPE_ANNE) {
					var srtRel=MEMG.lookupByTXID($rObj.sys_ref_txid);
					if (srtRel!=null) {
						if (srtRel.r==MEM.anj_rejected.temp_nid) return "This was a rejected request to add an anjucator to a neuron.";
						else if (srtRel.r==MEM.anj_accepted.temp_nid) return "This was an accepted request to add an anjucator to a neuron.";
					}
				}
				else if ($rObj.to_type==TO_TYPE_TXID) {
					if ($rObj.r==MEM.anj_rejected.temp_nid) return "This was a rejected request to be added as an anjucator to a neuron.";
					else if ($rObj.r==MEM.anj_accepted.temp_nid) return "This was an accepted request to be added as an anjucator to a neuron.";
				}
			}
			else return "This is a request to add an anjucator to a neuron.";
		}
		if ($rObj.reln==MEM.winner_payout.nid) return "This is a winning payout for the boosties bucket draw."
		if ($rObj.reln==MEM.hanai_in.nid) {
			if ($rObj.sys_ref_txid==undefined || $rObj.sys_ref_txid==null || $rObj.sys_ref_txid=="") {
				return "This is approval of a HANAI request.  This TX should be mined shortly and become official.";
			}
			return "This is approval of a HANAI request.";
		}
		if ($rObj.reln==MEM.hanai_in_reject.nid) {
			if ($rObj.sys_ref_txid==undefined || $rObj.sys_ref_txid==null || $rObj.sys_ref_txid=="") {
				return "This is rejection of a HANAI request.  This TX should be mined shortly and become official.";
			}
			return "This is rejection of a HANAI request.";
		}
        if ($rObj.r==MEM.andy_collab_accepted.temp_nid) return "This is an accepted collab request.";
        if ($rObj.r==MEM.andy_collab_rejected.temp_nid) return "This is a rejected collab request.";
        if ($rObj.r==MEM.hanai_out.temp_nid) {
            var whoNOBJ = MEMG.getNeuronByTempNid($rObj.t);
			return "This is a request to HANAI out a neuron into the <b>"+whoNOBJ.bestNameHTML()+"</b> worldview.";
        }
		if ($rObj.r==MEM.anndle_offer.temp_nid) {
			var toNOBJ = MEMG.getNeuronByTempNid($rObj.t);
			var pending = "is a pending";
			if (ANNE.UTILS.isset($rObj.sys_ref_txid)==true) pending = "was an ";
			return "This "+pending+" ANNDLE offer that paid "+UTILS.formatNumber($rObj.annies)+" annies ("+UTILS.formatNumber($rObj.annies/100)+" ANNE coin) to control the '"+toNOBJ.text+"' ANNDLE.";
		}

        if ($rObj.type==TYPE_SYS_MSG) {
			if ($rObj.r==MEM.annexed_by.temp_nid) {

				var arel = null;
				var ANNEXED_THING = null;

				if ($rObj.sys_ref_txid!=null) arel = ANNE.MEMG.lookupByTXID($rObj.sys_ref_txid);
				if (arel!=null) {
					ANNEXED_THING = MEMG.getNeuronByTempNid(arel.t);
				}

				if ($rObj.f==USER._userNObj.nid) {
					if (ANNEXED_THING!=null) {
						$rObj.msg = "Your neuron <b>"+ANNEXED_THING.bestName()+"</b> was <b>annexed by</b> "+MEMG.getOrMakeIfNew($rObj.to_s).bestName(); 
					}
					else {
						$rObj.msg = "Your neuron was <b>annexed by</b> "+MEMG.getOrMakeIfNew($rObj.to_s).bestName(); 
					}
				}
				else {
					if (ANNEXED_THING!=null) {
						$rObj.msg = "Neuron <b>"+ANNEXED_THING.bestName()+"</b> was <b>annexed by</b> "+MEMG.getOrMakeIfNew($rObj.to_s).bestName(); 
					}
					else {
						$rObj.msg = "A neuron from user "+MEMG.getOrMakeIfNew($rObj.to_s)+" was <b>annexed by</b> "+MEMG.getOrMakeIfNew($rObj.to_s).bestName(); 
					}
				}
			}

            //if ($rObj.r==MEM.aon.temp_nid) return "This is an onboarding relon with "+$rObj.annies+" annies given to the recipient.";
            return "This is a relon sent as a <b>message</b> to an ANNODE for instructions to take action.  It can also be generated by the annode on behalf of a keyless neuron and serve as a status update or receipt.";
        }

        else if ($rObj.type==TYPE_SYS) return "This is a <b>system</b> relon used to engage with ANNE protocol rules.";
        else if ($rObj.type==TYPE_BE) {
            if ($rObj.r==MEM.class.temp_nid || $rObj.r==MEM.inst.temp_nid) {
                return "This is a relon that specifies be-hierarchical parent information."+rel_sponsor_msg;
            }
            else if ($rObj.r==MEM.name.temp_nid) return "This is a relon that sets the neuron's name."+rel_sponsor_msg;
            else if ($rObj.r==MEM.aka.temp_nid) return "This is a relon that sets an <b>also known as</b> for this neuron."+rel_sponsor_msg;
            else if ($rObj.r==MEM.fka.temp_nid) return "This is a relon that sets a <b>formerly known as</b> for this neuron."+rel_sponsor_msg;
        	return "This is a relon that specifies additional be-defining information."+rel_sponsor_msg; 
        }
        else if ($rObj.type==TYPE_PARAM) {
            return "This is a <b>parameter</b> relon defining a relationship that instances can provide."+rel_sponsor_msg;
        }
        else if ($rObj.type==TYPE_HAS) {
            if (this.bCOLLAB==true) {
                if (this.bISPENDING==true || ($rObj.status_text!=undefined && $rObj.status_text=="p")) {
					return "This is a collab relon request.";
                }
                return "This is a collab relon request.";
            }
            return "This is a relon which specifies a parameter value a neuron <b>has</b>."+rel_sponsor_msg;
        }
        else if ($rObj.type==TYPE_AWARENESS) {
            return "This is a relon for run-time feelz and opinionz of an identity using ANNE.";
        }
        else if ($rObj.type==TYPE_BRAIN) {
            return "This is a relon used for wiring neurons together in the non-semantic <b>brain</b> dimension."+rel_sponsor_msg;
        }
        else if ($rObj.type==TYPE_CLUMP) return "This is a relon connecting to a more clump neuron."+rel_sponsor_msg;
        else if ($rObj.type==TYPE_ROLE) {
            return "This is a semantic role relon that exists in a clump neuron and defines a semantic role within the clump."+rel_sponsor_msg;
        }
		else if ($rObj.type==TYPE_INTERNAL_PAY) {
			if($rObj.f==ANNE.MEM.___SELF___.temp_nid && $rObj.r==MEM.BLOCK_WINNER.temp_nid) {
				var toNOBJ = MEMG.getNeuronByTempNid($rObj.t);
				return "This is a block reward won via solo mining by "+toNOBJ.text;
			}
			else if($rObj.f==ANNE.MEM.___SELF___.temp_nid && $rObj.r==MEM.BLOCK_SHARE_WINNER.temp_nid) {
				var toNOBJ = MEMG.getNeuronByTempNid($rObj.t);
				return "This is a block share reward won via share mining by "+toNOBJ.text;
			}
			else if ($rObj.r==ANNE.MEM.boosties.temp_nid) {
				return "This added boosties to this neuron.";
			}
            else return "This is an internal payment from a keyless neuron made by L1 protocol rules.";
        }
		else if ($rObj.type>=TYPE_BRAIN0 && $rObj.type<TYPE_BRAIN120) {
			return "This is a special type to indicate sorting order #"+$rObj.type;
		}
		else if($rObj.type==TYPE_BRAIN120) {
			return "This is a special BRAIN120 relon created by the ANNODE to document backlinks and provide helpful summaries.";
		}
		else if ($rObj.type==TYPE_BRAIN121) {
			return "myelination";
		}
	}


	logFunc=function($funct, $str="", $str2="", $str3="") {
		if ($str3!="") console.log("%c("+$funct+") %c"+$str+" %c"+$str2+" %c"+$str3, TALK.funct, TALK.bggreen, TALK.bgblue, TALK.bgyellow);
		else if ($str2!="") console.log("%c("+$funct+") %c"+$str+" %c"+$str2, TALK.funct, TALK.bggreen, TALK.bgblue);
		else if ($str!="") console.log("%c("+$funct+") %c"+$str, TALK.funct, TALK.bggreen);
	}
        
       
        

}


const TALK = new anne_talk ();
TALK.set_init(TALK);