/**
 * We will potentially have MILLIONS of relons in IDB memory.
 * In earlier iterations of the MEM we loaded ALL relons into the hypergraph
 * and had access anywhere in the app to the neurons/relons outs/ints etc.
 * 
 * This is not possible with a big memory footprint nor efficient.
 * 
 * We now load the relons from IDB on-demand, typically when we are going to check
 * ANNE for more relons.  This ensures we only load MEM in RAM when we need it.
 * 
 * We do a default MEM load of relons we store in the "relonsMinBoot" table everytime 
 * any app starts.  This will only grow with basic class definitions and some other
 * relons we target to store at bootup.
 */


var IDB = {}

//these pulls from IDB are NEVER updating IDB.  We just need to run them throught the MEM processor
//to get them added in to the local RAM hypergraph.
IDB.bUPDATEMEM=false;


/**
 * 
 * @param {*} $phObj 
 * @param {*} $nid 
 * @param {*} $bSPONSOREDTOO 
 * @returns 
 */
IDB.getAllMyStuff=async($phObj=null, $nid, $bSPONSOREDTOO)=> {
    if ($nid==undefined) {
        alert("FIX THIS JACKYCHAN");
        return [];
    }
    
    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    if ($phObj!=null) {
        if ($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }
    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];

    for (var r=0; r < aRelonsToProcess.length; r++) {
        if (aRelonsToProcess[r].from_nid!=$nid) {
            if (aRelonsToProcess[r].to_s!=$nid) {
                if ($bSPONSOREDTOO==false) continue;
                if (aRelonsToProcess[r].sponsored_by!=undefined && aRelonsToProcess[r].sponsored_by!=$nid) {
                    if (aRelonsToProcess[r].relon_sponsor!=undefined && aRelonsToProcess[r].relon_sponsor!=$nid) {
                        continue;
                    }
                }
            }
        }
        aRelons.push(aRelonsToProcess[r]);
    }
    
    console.log("%c IDB getAllMy$tuff RAM LOAD%c"+aRelons.length+"%c relons", TALK.bgyellow, TALK.bggreen, TALK.bgblack);
    
    if (aRelons.length==0) {
        return [];
    }

    //NOTE:when we remove the timer just return the return value and don't store it in aRelons2.
	var aRelons2 = await MEMP.processNeuronsFromRelons(aRelons, IDB.bUPDATEMEM, false, "relons", "<<from>> IDB.getAllMyStuff");

    return aRelons2;
}


/**
 * When querying for recents we pull a bunch of relons.
 * To avoid insert errors and to avoid looking up each txid in IDB one by one
 * We will load some memory here.
 * @param {*} $rels 
 * @returns 
 */
IDB.extractAndIDBFromNids=async($rels)=>{
    var nids = [];
    for (var r=0; r < $rels.length; r++) {
        if (nids.indexOf($rels[r].from_nid)==-1) {
            nids.push($rels[r].from_nid);
        }
    }
    var types=[0,1,2];
    var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.lookupNids", nids.join("")+types.join(""));
    await IDB.lookupNids(phObj, nids, types);
    return;
}


/**
 * use an array map of new data relons we have pulled from ANNE.
 * Some may be already in IDB memory but not in RAM.
 * 
 * Loop IDB memory to find what we have already stored in IDB.
 * If we have it in IDB, take it out of the relonsMap (but still process it into RAM)
 * 
 * The leftover .arr will be the new relons.
 *
 * @param {array_map} $relonsMap 
 */
IDB.checkForNewTxids=async($relonsMap)=>{
    
    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        var obj = $relonsMap.get(aRelonsToProcess[r].txid);
        if (obj!=null) {
            $relonsMap.remove(aRelonsToProcess[r].txid, "txid");
            aRelons.push(obj);
            continue;
        }
    }

    if (aRelons.length==0) return [];

    //only putting the net-diff aRelons in RAM here.
    await MEMP.processNeuronsFromRelons(aRelons, false, false, "relons", "IDB.checkForNewTxids");
}


/**
 * 
 * @param {*} $relonsMap 
 */
IDB.checkForNewTxids_ONLY=async($relonsMap, $_thisRunNECROSIS=[])=>{

    //not needed, we loaded ALL mem at start
    var a1 = await DB.loadAll("relons");
    var a2 = await DB.loadAll("relonsMinBoot");
    
    var aRelonsToProcess = a1.concat(a2);

    var rs = [];

    //process ALL RELONS
    for (var r=0; r < aRelonsToProcess.length; r++) {

        if ($relonsMap.arr.length==0) return;
            
        //in this data pull we found a neuron that is nuked.
        //tag ALL relons of that neuron as nuked.
        //we will remove them from IDB and RAM.
        if ($_thisRunNECROSIS.length) {
            if (aRelonsToProcess[r].to_type==TO_TYPE_ANNE && $_thisRunNECROSIS.indexOf(aRelonsToProcess[r].to_s)!=-1) {
                if (MEM._localRunDelete.indexOf(aRelonsToProcess[r].txid)===1) MEM._localRunDelete.push(aRelonsToProcess[r].txid);
                var obj = $relonsMap.get(aRelonsToProcess[r].txid);
                if (obj!=null) {
                    obj.idb_insert=false;
                    obj.idb_update=false;
                    obj.idb_delete=true;
                    obj.active=false;
                    obj.bad_data=true;
                }
                continue; 
            }

            else if ($_thisRunNECROSIS.indexOf(aRelonsToProcess[r].reln)!=-1) {
                if (MEM._localRunDelete.indexOf(aRelonsToProcess[r].txid)===1) MEM._localRunDelete.push(aRelonsToProcess[r].txid)
                var obj = $relonsMap.get(aRelonsToProcess[r].txid);
                if (obj!=null) {
                    obj.idb_insert=false;
                    obj.idb_update=false;
                    obj.idb_delete=true;
                    obj.active=false;
                    obj.bad_data=true;
                }    
                continue;     
            }

            else if ($_thisRunNECROSIS.indexOf(aRelonsToProcess[r].from_nid)!=-1) {
                if (MEM._localRunDelete.indexOf(aRelonsToProcess[r].txid)===1) MEM._localRunDelete.push(aRelonsToProcess[r].txid)
                var obj = $relonsMap.get(aRelonsToProcess[r].txid);
                if (obj!=null) {
                    obj.idb_insert=false;
                    obj.idb_update=false;
                    obj.idb_delete=true;
                    obj.active=false;
                    obj.bad_data=true;
                }
                continue;        
            }
        }

        //the idb relon we currently have in IDB is in our list of relons we are about to process.
        var obj = $relonsMap.get(aRelonsToProcess[r].txid);

        //we have it already in IDB mem..
        // && (obj.force_update==undefined || obj.force_update==false)
        if (obj!=null && obj.bMUSTINSERT==undefined) {
            
            obj.idb_insert = false;
            obj.idb_update = true;

            //remove from the .arr array --- WE HAVE THIS IN IDB MEMORY.
            $relonsMap.remove(aRelonsToProcess[r].txid, "txid");
            $relonsMap.arr_other.push(obj);

             //unconf shoudl be -1 now (NOT 0);

            if (obj.time!=undefined && (obj.temp==undefined || obj.temp==false)) {
                let time = parseInt(obj.time);
                if (time > HISTORY._GREATEST_LAST_KNOWN_BLOCK) HISTORY.updateGLKB_fromRelons(time);
            }

            if (obj.bMUSTINSERT!=undefined) {
                obj.dupe=false;
                obj.temp=false;
                obj.idb_insert=true;
                obj.idb_update=false
                delete obj.bMUSTINSERT;
                continue;
            }

            var dupe_robj_in_RAM = ANNE.MEMG.lookupByTXID(obj.txid);
            var chgs=0;


            //relon in mem is a temp, but we have a REAL relon now.
            if (dupe_robj_in_RAM!=null && dupe_robj_in_RAM.temp==true && obj.ghost==false) {
                if (obj.time!=undefined && obj.time==-1) {
                    if (obj.txid[0]=="z") {
                        obj.ghost=true;
                    }
                    if (bNESTED==false && ANNE.MEM_UDF.bGETDEDSATBOOTUP==true) {
                        _memPoolBootups.push(rObj.txid);
                    }
                }
                else {
                    dupe_robj_in_RAM.temp=false;
                }
            }
            

            //parent is different than what we have in IDB
            if ((obj.parent!=undefined && obj.parent!=null)
                && ((aRelonsToProcess[r].parent==undefined || aRelonsToProcess[r].parent==null)
                || obj.time!=aRelonsToProcess[r].parent)) {
                if (dupe_robj_in_RAM!=null) dupe_robj_in_RAM.parent = obj.parent;
                chgs+=1;
            }

            //time is different than what we have in IDB
            if ((obj.time!=undefined && obj.time!=null)
                && ((aRelonsToProcess[r].time==undefined || aRelonsToProcess[r].time==null)
                || obj.time!=aRelonsToProcess[r].time)) {
                if (dupe_robj_in_RAM!=null) dupe_robj_in_RAM.time = obj.time;
                chgs+=1;
            }
            

            //sponsored_by is different than what we have in IDB
            if ((obj.sponsored_by!=undefined && obj.sponsored_by!=null)
                &&  ((aRelonsToProcess[r].sponsored_by==undefined || aRelonsToProcess[r].sponsored_by==null)
                || obj.sponsored_by!=aRelonsToProcess[r].sponsored_by)) {
                if (dupe_robj_in_RAM!=null) {
                    if (obj.from_nidTNOBJ==null) obj.from_nidTNOBJ = MEMG.getOrMakeIfNew(obj.from_nid);
                    dupe_robj_in_RAM.sponsored_by = obj.sponsored_by;
                    if (dupe_robj_in_RAM.sponsored_by!=undefined && dupe_robj_in_RAM.sponsored_by!="") {
                        let oldSP = MEMG.getOrMakeIfNew(dupe_robj_in_RAM.sponsored_by);
                        if (oldSP) obj.from_nidTNOBJ._removeSponsoredBy(oldSP.temp_nid);
                    }
                }
                chgs+=1;
            }

            //to_s is different than what we have in IDB
            if ((obj.to_s!=undefined && obj.to_s!=null)
                &&  ((aRelonsToProcess[r].to_s==undefined || aRelonsToProcess[r].to_s==null)
                || obj.to_s!=aRelonsToProcess[r].to_s)) {
                if (dupe_robj_in_RAM!=null) {
                    dupe_robj_in_RAM.to_s = obj.to_s;
                    if (obj.to_sTNOBJ==null) obj.to_sTNOBJ = MEMG.getNeuronByNid(obj.to_s);
                    if (obj.to_sTNOBJ!=null) {
                        obj.t = obj.to_sTNOBJ.temp_nid;
                        dupe_robj_in_RAM.t = obj.to_sTNOBJ.temp_nid;
                    }        
                }
                chgs+=1;
            }


            //the to_s will change on a vote (we create the annode vote relon before they vote)..
            //we tie it to the unvoted to_s but also the ref_txid pts to their anjucator relon
            //in the source neuron.  That sys_ref_txid is updated to their actual on-chain vote
            //when they do infact vote.
            if (dupe_robj_in_RAM!=null) {
                if ((dupe_robj_in_RAM.to_s==MEM.anj_vote_hasnot.nid) && ((obj.to_s!=MEM.anj_vote_hasnot.nid))) {      
                    dupe_robj_in_RAM.to_s=obj.to_s;
                    if (obj.to_sTNOBJ==null) obj.to_sTNOBJ = MEMG.getNeuronByNid(rObj.to_s);
                    if (obj.to_sTNOBJ!=null) {
                        dupe_robj_in_RAM.t = obj.to_sTNOBJ.temp_nid;
                        obj.t = obj.to_sTNOBJ.temp_nid;
                    }
                    chgs+=1;
                }
            }


            //owned_by is different than what we have in IDB
            if ((obj.owned_by!=undefined && obj.owned_by!=null)
                &&  ((aRelonsToProcess[r].owned_by==undefined || aRelonsToProcess[r].owned_by==null)
                || obj.owned_by!=aRelonsToProcess[r].owned_by)) {
                if (dupe_robj_in_RAM!=null) {
                    dupe_robj_in_RAM.owned_by = obj.owned_by;
                    if (obj.from_nidTNOBJ==null) obj.from_nidTNOBJ = MEMG.getOrMakeIfNew(obj.from_nid);
                        if (dupe_robj_in_RAM.owned_by!=undefined && dupe_robj_in_RAM.owned_by!="") {
                            let oldOw = MEMG.getOrMakeIfNew(dupe_robj_in_RAM.owned_by);
                            if (oldOw) obj.from_nidTNOBJ._removeOwns(oldOw.temp_nid);
                        }
                        dupe_robj_in_RAM.owned_by=obj.owned_by;
                        obj.from_nidTNOBJ.owned_by=obj.owned_by;
                    }   
                chgs+=1;
            }

            //relon_sponsor is different than what we have in IDB
            if ((obj.relon_sponsor!=undefined && obj.relon_sponsor!=null)
                &&  ((aRelonsToProcess[r].relon_sponsor==undefined || aRelonsToProcess[r].relon_sponsor==null)
                || obj.relon_sponsor!=aRelonsToProcess[r].relon_sponsor)) {
                if (dupe_robj_in_RAM!=null) {
                    dupe_robj_in_RAM.relon_sponsor = obj.relon_sponsor;
                    dupe_robj_in_RAM.relon_sponsor=obj.relon_sponsor;
                    if (rs.indexOf(dupe_robj_in_RAM.relon_sponsor)==-1) {
                        ANNE.MEMG.getOrMakeIfNew(dupe_robj_in_RAM.relon_sponsor, "", NTYPE_UNKNOWN, 0);
                        rs.push(dupe_robj_in_RAM.relon_sponsor);
                    }
                    //add to our tracking array.
                    if (ANNE.USER._userNObj!=null && obj.relon_sponsor==ANNE.USER._userNObj.nid) {
                        ANNE.USER._userNObj.addUsAsRelonSponsor(obj);
                    }
                }
                chgs+=1;
            }

            //first seen is different than what we have in IDB
            if ((obj.firstseen!=undefined && obj.firstseen!=null)
                &&  ((aRelonsToProcess[r].firstseen==undefined || aRelonsToProcess[r].firstseen==null)
                || obj.firstseen!=aRelonsToProcess[r].firstseen)) {
                if (dupe_robj_in_RAM!=null) dupe_robj_in_RAM.firstseen = obj.firstseen;
                chgs+=1;
            }

            //annies are different than what we have in IDB
            if ((obj.annies!=undefined && obj.annies!=null)
                &&  ((aRelonsToProcess[r].annies==undefined || aRelonsToProcess[r].annies==null)
                || obj.annies!=aRelonsToProcess[r].annies)) {
                if (dupe_robj_in_RAM!=null) dupe_robj_in_RAM.annies = obj.annies;
                chgs+=1;
            }

            //freeze is different than what we have in IDB
            if ((obj.freeze!=undefined && obj.freeze!=null)
                &&  ((aRelonsToProcess[r].freeze==undefined || aRelonsToProcess[r].freeze==null)
                || obj.freeze!=aRelonsToProcess[r].freeze)) {
                if (dupe_robj_in_RAM!=null) dupe_robj_in_RAM.freeze = obj.freeze;
                chgs+=1;
            }

            //sys ref txid is different than what we have in IDB
            if ((obj.sys_ref_txid!=undefined && obj.sys_ref_txid!=null)
                &&  ((aRelonsToProcess[r].sys_ref_txid==undefined || aRelonsToProcess[r].sys_ref_txid==null)
                || obj.sys_ref_txid!=aRelonsToProcess[r].sys_ref_txid)) {
                if (dupe_robj_in_RAM!=null) {
                    dupe_robj_in_RAM.sys_ref_txid = obj.sys_ref_txid;
                    dupe_robj_in_RAM.to_s = obj.to_s;
                    if (obj.to_sTNOBJ==null) obj.to_sTNOBJ = MEMG.getNeuronByNid(obj.to_s);
                    if (obj.to_sTNOBJ!=null) {
                        dupe_robj_in_RAM.t = obj.to_sTNOBJ.temp_nid;
                        obj.t = obj.to_sTNOBJ.temp_nid;
                    }
                }
                chgs+=1;
            }

            if (chgs>0 && dupe_robj_in_RAM!=null || (obj.force_update!=undefined && obj.force_update==true)) {
                obj.dupe_robj_in_mem = dupe_robj_in_RAM;
                obj.idb_update = true;
                obj.dupe=true;
                delete obj.force_update;
                continue;
            }

            if (obj.relon_sponsor!=undefined) {
                if (ANNE.USER._userNObj!=null && obj.relon_sponsor==ANNE.USER._userNObj.nid) {
                    ANNE.USER._userNObj.addUsAsRelonSponsor(obj);
                }
            }

            obj.idb_update = false;
            continue;
        }
    }
    return;
}


/**
 * @param {*} $phObj 
 * @param {*} $nid 
 * @returns 
 */
IDB.getAllPayments = async($phObj=null, $nid) =>{
    if ($nid==undefined) {
        alert("FIX THIS JACKYCHAN");
        return [];
    }

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }
    
    if ($phObj!=null) {
        if ($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }
    
    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if (aRelonsToProcess[r].from_nid!=$nid && aRelonsToProcess[r].to_s!=$nid) continue;
        if (aRelonsToProcess[r].type!=TYPE_INTERNAL_PAY && aRelonsToProcess[r].typ!=TYPE_BROADCAST_PAY) {
            if ((aRelonsToProcess[r].annies==undefined || aRelonsToProcess[r].annies==null || aRelonsToProcess[r].annies==0)) continue;
        }
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) return [];

    var aRelons2 = await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.getAllPayments");
    return aRelons2;
}



/**
 * 
 * @param {*} $phObj 
 * @param {*} $nid 
 * @param {*} $types 
 * @returns 
 */
IDB.queryForOuts_not = async($phObj=null, $nid, $types)=>{
    if ($nid==undefined) {
        alert("FIX THIS JACKYCHAN");
        return [];
    }
    
    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    if ($phObj!=null) {
        if ($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }

    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if (aRelonsToProcess[r].from_nid!=$nid) continue;

        //we WANT outs NOT in the $types array.  Skip those in the types array.
        if ($types.indexOf(aRelonsToProcess[r].type)!=-1) continue;
        
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) return [];

   
    var aRelons2 = await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.qqueryForOuts_not");
    return aRelons2;
}


/**
 * 
 * @param {*} $phObj 
 * @param {*} $nids 
 * @param {*} $types 
 * @returns 
 */
IDB.queryForOutsFroms_not = async($phObj=null, $nids, $types=[])=>{
    if ($nids==undefined) {
        alert("FIX THIS JACKYCHAN");
        return [];
    }

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }
    
    if ($phObj!=null) {
        if ($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }

    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if ($nids.indexOf(aRelonsToProcess[r].from_nid)==-1) {
            continue;
        }
        if ($types.length>0 && $types.indexOf(aRelonsToProcess[r].type)!=-1) continue;
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) return [];
    var aRelons2 = await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.queryForOutsFroms_not");	
    return aRelons2;
}


/**
 * 
 * @param {*} $phObj 
 * @param {*} $nid 
 * @param {*} $types 
 * @returns 
 */
IDB.queryForOuts=async($phObj=null, $nid, $types=[])=>{
    if ($nid==undefined) {
        alert("FIX THIS JACKYCHAN");
        return [];
    }
    
    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    if ($phObj!=null) {
        if ($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }
    
    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if (aRelonsToProcess[r].from_nid!=$nid) {
            continue;
        }
        if ($types.length>0 && $types.indexOf(aRelonsToProcess[r].type)==-1) continue;
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) return [];

    var aRelons2=[];
    if (aRelons.length) {
        aRelons2 = await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.queryForOuts");	
    }
    return aRelons2;
}


/**
 * 
 * @param {*} $phObj 
 * @param {*} $nid 
 * @param {*} $types 
 * @returns 
 */
IDB.queryForIns=async($phObj=null, $nid)=>{
    if ($nid==undefined) {
        alert("FIX THIS JACKYCHAN");
        return [];
    }

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }
    if ($phObj!=null) {
        if ($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }

    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    var beFroms = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if (aRelonsToProcess[r].to_s!=$nid) {
            continue;
        }
        aRelons.push(aRelonsToProcess[r]);
        
        //get from nids (we atleast need names in RAM)
        if (beFroms.indexOf(aRelonsToProcess[r].from_nid)==-1) {
            beFroms.push(aRelonsToProcess[r].from_nid)
        }
    }

    for (var r=0; r < aRelonsToProcess.length; r++) {
        if (aRelonsToProcess[r].type!=TYPE_BE) continue;
        if (beFroms.indexOf(aRelonsToProcess[r].from_nid)==-1) continue;
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) return [];
    var aRelons2=[];
    if (aRelons.length) {
        aRelons2 = await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.queryForIns");	
    }
    return aRelons2.concat(aRelons);
}


/**
 * 
 * @param {*} $phObj 
 * @param {*} $nid 
 * @param {*} $types 
 * @returns 
 */
IDB.queryForOutsAndIns=async($phObj=null, $nid, $types=[])=>{
    if ($nid==undefined) {
        alert("FIX THIS JACKYCHAN");
        return [];
    }

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }
    
    if ($phObj!=null) {
        if ($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }
    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if (aRelonsToProcess[r].from_nid!=$nid) {
            if (aRelonsToProcess[r].to_s!=$nid) {
                continue;
            }
        }
        if ($types.length>0 && $types.indexOf(aRelonsToProcess[r].type)==-1) continue;
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) return [];
    var aRelons2 = await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.queryForOutsAndIns");	
    return aRelons2;
}


IDB.queryForOutsAndInsFroms=async($phObj=null, $nids, $types=[])=>{
    if ($nids==undefined) {
        alert("FIX THIS JACKYCHAN");
        return [];
    }
    
    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    if ($phObj!=null) {
        if ($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }
    
    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if ($types.length>0 && $types.indexOf(aRelonsToProcess[r].type)==-1) continue;
        
        if ($nids.indexOf(aRelonsToProcess[r].from_nid)==-1) {
            if ($nids.indexOf(aRelonsToProcess[r].to_s)==-1) {
                continue;
            }
        }
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) return [];
    var aRelons2 = await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "queryForOutsAndForInsFroms");	
    return aRelons2;
}


/**
 * 
 * @param {*} $phObj 
 * @param {*} $bem_nids 
 * @param {*} $types 
 * @returns 
 */
IDB.queryForOutsAndIns_BEM=async($phObj=null, $bem_nids, $types=[])=>{
    if ($bem_nids==undefined) {
        alert("FIX THIS JACKYCHAN");
        return [];
    }

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }
    
    if ($phObj!=null) {
        if ($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }
    
    if ($bem_nids.length) {
        var aRelonsToProcess = await DB.loadAll("relons");
        var aRelons = [];
        for (var r=0; r < aRelonsToProcess.length; r++) {
            if ($types.length>0 && $types.indexOf(aRelonsToProcess[r].type)==-1) continue;
            if ($bem_nids.indexOf(aRelonsToProcess[r].from_nid)==-1) continue;
            aRelons.push(aRelonsToProcess[r]);
        }

        if (aRelons.length==0) return [];
        aRelons = await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.queryForOutsAndIns_BEM");
        return aRelons;
    }
}


/**
 * 
 * @param {*} $phObj 
 * @param {*} $relns 
 * @returns 
 */
IDB.queryTosFromRelns=async($phObj=null, $relns)=>{
    if ($relns==undefined) {
        alert("FIX THIS JACKYCHAN");
        return [];
    }
    
    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    if ($phObj!=null) {
        if ($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }

    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    var nids = [];

    for (var r=0; r < aRelonsToProcess.length; r++) {
        if (aRelonsToProcess[r].to_type!=TO_TYPE_ANNE) continue;
        if ($relns.indexOf(aRelonsToProcess[r].reln)==-1) {
            if ($relns.indexOf(aRelonsToProcess[r].to_s)==-1 || aRelonsToProcess[r].reln!=MEM.ec_is.nid) {
                continue;
            }
            nids.push(aRelonsToProcess[r].to_s);
        }
        else {
            nids.push(aRelonsToProcess[r].from_nid);
        }

        if (nids.indexOf(aRelonsToProcess[r].reln)!=-1) continue;
        
    }

    //get all those outs.
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if (aRelonsToProcess[r].type==TYPE_AWARENESS) continue;
        if (nids.indexOf(aRelonsToProcess[r].from_nid)==-1) continue;
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) return [];
    aRelons = await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.queryTosFromRelons");
    return aRelons;
}


/**
 * This only gets the direct parent..
 * Not BEM and not the creative alt-be parent lines.
 * @param {*} $phObj 
 * @param {*} $parent_nid 
 * @returns 
 */
IDB.getChldren=async($phObj=null, $parent_nid, $ntype=null)=>{
    if ($parent_nid==undefined) {
        alert("FIX THIS JACKYCHAN");
        return [];
    }
    
    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    if ($phObj!=null) {
        if ($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }
    
    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if ($ntype!=null && aRelonsToProcess[r].ntype!=$ntype) continue;
        if (aRelonsToProcess[r].type!=TYPE_BE) continue;
        if (aRelonsToProcess[r].to_type!=TO_TYPE_ANNE) continue;
        if (aRelonsToProcess[r].to_s!=$parent_nid) continue;
        if (aRelonsToProcess[r].r!=MEM.inst.temp_nid && aRelonsToProcess[r].r!=MEM.class.temp_nid) continue;
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) return [];

    aRelons = await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.getChildren");
    return aRelons;
}


/**
 * 
 * @param {*} $phObj 
 * @param {*} $bem_nids 
 * @param {*} $types 
 * @returns 
 */
IDB.queryForOuts_BEM=async($phObj=null, $bem_nids, $types=[])=>{
    if ($bem_nids==undefined) {
        alert("FIX THIS JACKYCHAN");
        return [];
    }

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }
    
    if ($phObj!=null) {
        if ($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }
    
    if ($bem_nids.length==0) return [];

    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if ($types.length>0 && $types.indexOf(aRelonsToProcess[r].type)==-1) continue;
        if ($bem_nids.indexOf(aRelonsToProcess[r].from_nid)==-1) {
            continue;
        }
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) return [];

    aRelons = await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.queryForOuts_BEM");
    return aRelons;
}


/**
 * 
 * @param {*} $phObj 
 * @param {*} $nid 
 * @returns 
 */
IDB.queryForCircuitOutsAndIns=async($phObj=null, $nid)=>{
    if ($nid==undefined) {
        alert("FIX THIS JACKYCHAN");
        return [];
    }

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }
    
    if ($phObj!=null) {
        if ($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }
    
    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if (aRelonsToProcess[r].from_nid!=$nid) {
            if (aRelonsToProcess[r].to_s!=$nid )continue;
        }
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) return [];
    aRelons = await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.queryForCircuitOutsAndIns");
    return aRelons;
}


/**
 * 
 * @param {*} $phObj 
 * @param {*} $text 
 * @param {*} $class_nid 
 * @param {*} $wvs 
 * @returns 
 */
IDB.queryForText_BEM=async($phObj=null, $text, $class_nid, $wvs)=>{
    
    if ($phObj!=null) {
        if ($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    var lowerText = $text.toLowerCase();
    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    var classNOBJ = MEMG.getNeuronByNid($class_nid);
    if (classNOBJ==null) return [];

    for (var r=0; r < aRelonsToProcess.length; r++) {
        if ($wvs.length && $wvs.indexOf(aRelonsToProcess[r].owned_by)==-1) continue;
        
        if (aRelonsToProcess[r].to_type==TO_TYPE_STRING && aRelonsToProcess[r].to_s.toLowerCase()!=lowerText) continue;
        if (aRelonsToProcess[r].text!=undefined && aRelonsToProcess[r].text.toLowerCase()!=lowerText) continue;
         
        if (MEMG.isBesOfs(classNOBJ, aRelonsToProcess[r].bes)==false) continue;
        if (classNOBJ.is()==false) continue;
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) return [];
    await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.queryForTExt_BEM");
    return aRelons;
}


/**
 * 
 * @param {*} $text 
 * @param {*} $phObj 
 * @param {Boolean} $bEXACT
 * @returns 
 */
IDB.queryForText=async($text, $phObj=null, $bEXACT=true)=>{
    if ($text==undefined || $text==null || typeof $text!=="string") {
        console.warn("fix this .. passed no or wrong var to IDB.queryForText");
        return [];
    }

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    if ($phObj==null) {
        $phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForText", $text+$bEXACT+"");
        $phObj.bDIDIDB=true;
    }
    else if ($phObj.bDIDIDB==true) return [];
   
    var lowerText = $text.toLowerCase();
    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];

    //this will get relons of names and aliases from IDB.  
    if ($bEXACT==true) {
        for (var r=0; r < aRelonsToProcess.length; r++) {
            if (aRelonsToProcess[r].to_type==TO_TYPE_STRING) {
                if(aRelonsToProcess[r].to_s.toLowerCase()==lowerText) {
                    aRelons.push(aRelonsToProcess[r]);
                    continue;
                }
            }
            else if (aRelonsToProcess[r].text!=undefined && aRelonsToProcess[r].text.toLowerCase()==lowerText) {
                aRelons.push(aRelonsToProcess[r]);
                continue;
            }
        }
    }

    else {
        for (var r=0; r < aRelonsToProcess.length; r++) {
            if (aRelonsToProcess[r].to_type==TO_TYPE_STRING) {
                if(aRelonsToProcess[r].to_s.toLowerCase().indexOf(lowerText)!=-1) {
                    aRelons.push(aRelonsToProcess[r]);
                    continue;
                }
            }
            else if (aRelonsToProcess[r].text!=undefined && aRelonsToProcess[r].text.toLowerCase().indexOf(lowerText)!=-1) {
                aRelons.push(aRelonsToProcess[r]);
                continue;
            }
        }
    }

    if (aRelons.length==0) return [];
 
    if (aRelons.length>1000) {
        console.log("audit spot check this..");
        console.log("should '"+$text+"' be returning so many variables?");
    }
    await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.queryForText");
    return aRelons;
}


/**
 * 
 * @param {*} $phObj 
 * @param {*} $text 
 * @param {*} $be_nid 
 * @param {*} $aWVs 
 * @returns 
 */
IDB.queryForTextAndBe=async($phObj=null, $text, $be_nid, $WVs)=>{
    
    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    if ($phObj!=null) {
        if ($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }
    
    var lowerText = $text.toLowerCase();
    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if (aRelonsToProcess[r].be!=$be_nid) continue;
        if (aRelonsToProcess[r].from_text.toLowerCase()!=lowerText) continue;
        if ($WVs.length && $WVs.indexOf(aRelonsToProcess[r].owned_by)==-1) continue;
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) return [];
    await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.queryForTextAndBe");
    return aRelons;
}


IDB.queryForNid=async($phObj=null, $nid, $types=[])=>{
    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    return await IDB.lookupNid($phObj=null, $nid, $types);
}

IDB.getNid=async($phObj=null, $nid, $types=[])=>{
    
    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    return await IDB.lookupNid($phObj=null, $nid, $types);
}

/**
 * 
 * @param {*} $phObj 
 * @param {*} $nid 
 * @param {*} $types 
 * @returns 
 */
IDB.lookupNid=async($phObj=null, $nid, $types=[])=>{

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    if ($nid==undefined) {
        alert("FIX THIS JACKYCHAN");
        return [];
    }
    
    if ($phObj!=null) {
        if($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }

    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if (aRelonsToProcess[r].from_nid!=$nid) {
            continue;
        }
        if ($types.length && $types.indexOf(aRelonsToProcess[r].type)==-1) continue;
        
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) {
        return [];
    }

    var aRelons = await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.lookupNid");
    return aRelons;
}


/**
 * 
 * @param {*} $phObj 
 * @param {*} $nids 
 * @param {*} $types 
 * @param {*} $sysMsgRelns 
 * @returns 
 */
IDB.lookupNidsAndSysMsgRelns=async($phObj=null, $nids, $types=[], $sysMsgRelns)=>{
    if ($nids==undefined) {
        alert("FIX THIS JACKYCHAN");
        return [];
    }

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    if ($phObj!=null) {
        if ($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }
    
    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if ($nids.indexOf(aRelonsToProcess[r].from_nid)==-1) continue;
        if ($sysMsgRelns.length>0 && $sysMsgRelns.indexOf(aRelonsToProcess[r])==-1) continue;
        if ($types.length>0 && $types.indexOf(aRelonsToProcess[r].type)==-1) continue;
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) return [];

    var aRelons = await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.lookupNidsAndSysMsgRelns");
    return aRelons;
}


/**
 * 
 * @param {*} $phObj 
 * @param {array} $nids 
 * @param {array} $types
 * @param {string} $from_msg
 * @returns 
 */
IDB.lookupNids=async($phObj=null, $nids, $types=[], $from_msg="")=>{
    if ($nids==undefined) {
        alert("FIX THIS JACKYCHAN");
        return [];
    }

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    if($phObj==null) {
        $phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.lookupNids", $nids.join("")+$types.join(""));
    }
    if ($phObj!=null) {
        if ($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }
    
    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if ($nids.indexOf(aRelonsToProcess[r].from_nid)==-1) {
            continue;
        };
        if ($types.length>0 && $types.indexOf(aRelonsToProcess[r].type)==-1) continue;
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) return [];

    var aRelons = await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.lookupNids "+$from_msg);
    return aRelons;
}


/** 
 * 
 * @param {*} $phObj 
 * @param {*} $nids 
 * @returns 
 */
IDB.lookupNidsGetAllOuts=async($phObj=null, $nids)=>{
    if ($nids==undefined) {
        alert("FIX THIS JACKYCHAN");
        return [];
    }

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    if ($phObj==null) $phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.lookupNidsGetAllOuts ", $nids);
		
    if ($phObj!=null) {
        if ($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }
    
    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if ($nids.indexOf(aRelonsToProcess[r].from_nid)==-1) {
            continue;
        }
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) return [];

    var aRelons = await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.lookupNidsGetAllOuts");
    return aRelons;
}


/**
 * 
 * @param {*} $phObj 
 * @param {*} $type 
 * @param {*} $reln 
 * @param {*} $to 
 * @param {*} $to_type 
 * @returns 
 */
IDB.queryForInsWithTo=async($phObj=null, $type, $reln, $to, $to_type)=>{

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    if ($phObj!=null) {
        if ($phObj.bDIDIDB==true) return [];
        $phObj.bDIDIDB=true;
    }
    
    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if (aRelonsToProcess[r].to_s!=$to) continue;
        if (aRelonsToProcess[r].reln!=$reln) continue;
        if (aRelonsToProcess[r].type!=$type) continue;
        if (aRelonsToProcess[r].to_type!=$to_type) continue;
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) return [];
  
    var aRelons = await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.queryForInsWithTo");
    return aRelons;
}


/**
 * Targetted RAM load of specific nids and outs.
 * In our optimized pre-load we only take the firstseen X number of relons.
 * Any given app/use case may need to preload specific nid classes that come later.
 * @param {Array} $nids 
 */
IDB.loadIDBNidOuts=async($nids)=>{

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if ($nids.indexOf(aRelonsToProcess[r].from_nid)==-1) continue;
        aRelons.push(aRelonsToProcess[r]);
    }
    
    if (aRelons.length==0) return [];

    
    aRelons = await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.loadIDBNidOuts");
    return aRelons;
}


/**
 * 
 * @param {*} $txids 
 */
IDB.lookupTxids = async($txids)=>{

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if ($txids.indexOf(aRelonsToProcess[r].txid)==-1) continue;
        aRelons.push(aRelonsToProcess[r]);
        if ($txids.length==1) break;    //mini-time save, we have what we came for.
    }
    if (aRelons.length==0) return [];
    await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.lookupTxids");
}


/**
 * 
 * @param {*} $nid 
 * @param {*} $skip_type 
 */
IDB.loadIDBNidOuts_single=async($nid, $skip_type=TYPE_AWARENESS)=>{

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if (aRelonsToProcess[r].type==$skip_type) continue;
        if (aRelonsToProcess[r].from_nid!=$nid) continue;
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) return [];
    await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.loadIDBNidOuts_single");
}



/**
 * Targeted load of specific ins to nids     
 * In our optimized pre-load we only take the firstseen X number of relons.
 * Any given app/use case may need to preload specific nid classes that come later.
 * @param {Array} $nids 
 */
IDB.loadIDBNidIns=async($nids)=>{

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if ($nids.indexOf(aRelonsToProcess[r].to_s==-1)) continue;
        aRelons.push(aRelonsToProcess[r]);
    }

    if (aRelons.length==0) return [];
    await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.loadIDBNidIns");
}

IDB.loadIDBNidIns_single=async($nid)=>{

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.loadIDBNidIns_single ", $nid);
    if (phObj!=null) {
        if (phObj.bDIDIDB==true) return [];
        phObj.bDIDIDB=true;
    }
    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if ($nid!=aRelonsToProcess[r].to_s) continue;
        aRelons.push(aRelonsToProcess[r]);
    }
    if (aRelons.length==0) return [];
    await MEMP.processNeuronsFromRelons(aRelons, false, IDB.bUPDATEMEM, "relons", "IDB.loadIDBNidIns");
}


/**
 * 
 * @param {*} $nid 
 */
IDB.testForNid=async($nid)=> {

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }

    var yep = [];
    var rels = await DB.loadAll("relons");
    for (var r=0; r < rels.length; r++) {
        if (rels[r].from_nid==$nid) {
            yep.push(rels[r]);
        }
    }
    console.log("in relons table", yep);
    ANNE.TALK.printRelons(yep);
    var yep2 = [];
    var rels2 = await DB.loadAll("relonsMinBoot");
    for (var r=0; r < rels2.length; r++) {
        if (rels2[r].from_nid==$nid) {
            yep2.push(rels2[r]);
        }
    }
    console.log("in relons min boot table", yep2);
    ANNE.TALK.printRelons(yep2);
}


/**
 * Fix IDB memory after finding out run-time (in TX broadcast)
 * that a neuron has been updated with a owner/sponsor.
 * 
 * This comes from a broadcast hanai error message that informs us the 
 * neuron is already owned by ANNE.  (meaning a previous hanai already went through and the GUI is wrong)
 * 
 * @param {*} $nid 
 * @param {*} $owned_by 
 * @param {*} $sponsored_by 
 * @returns 
 */
IDB.updateAllOwnedBySponsoredBy=async($nid, $owned_by, $sponsored_by)=> {

    //not needed, we loaded ALL mem at start
    if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) { console.log("skip IDB, all loaded already"); return []; }
    
    var aRelonsToProcess = await DB.loadAll("relons");
    var aRelons = [];
    for (var r=0; r < aRelonsToProcess.length; r++) {
        if (aRelonsToProcess[r].from_nid!=$nid) continue;

        if (aRelonsToProcess[r].owned_by==undefined 
            || aRelonsToProcess[r].owned_by==null 
            || aRelonsToProcess[r].owned_by=="") continue;
        
        aRelonsToProcess[r].owned_by = $owned_by;
        aRelonsToProcess[r].sponsored_by = $sponsored_by;
        aRelons.push(aRelonsToProcess[r]);
    }

    await ANNE.DB.___updateAll(aRelons);
    return;
}