var HANDLER = {};

/*  
    Handlers are intended to handle ANNE classes of information. 
    An image handler would be capable of doing something with images (showing them) 
    
    For max use and sustainability,
    external handlers will need to be able to be dynamically defined, with ANNE TXs,
    in either seqs, or even ACQL circuits. 
    
    This was used in the MM project with the TMDB API circuit.  Neurons were defined that
    mapped their stuff to our stuff so an external json object could be parsed and conceptualized
    into ANNE's 1SCHEMA.
*/

HANDLER.init = async ()=> {
    var MEM = ANNE.MEM;

    if (MEM==undefined) {
        console.warn("MEM object must be defined first");
        return;
    }

    MEM.protocol = MEM.preloadWithOwner(MEM.ANNE,{text: "protocol", nid:'7771066892361241618'}, "class",  MEM.class);
    MEM.protocol_handler = MEM.preloadWithOwner(MEM.ANNE,{text: "protocol handler", nid:'777837351963390863'}, "class",  MEM.handler);
     
    //protocols expect to have a handler of class protocol handler
    MEM.preLoadFullRelon(TYPE_PARAM, MEM.protocol, MEM.handler, MEM.protocol_handler);
    
    MEM.external_protocol = MEM.preloadWithOwner(MEM.ANNE,{text: "external protocol", nid:'7779929479425985587'}, "class",  MEM.protocol);
    MEM.protocol_param = MEM.preloadWithOwner(MEM.ANNE,{text: "protocol param", nid:'7775956763254458021'}, "class",  MEM.__PARAM);
    
    //these protocols come w/ a spec of a specific sequence to order their protocol.
    //ie, [0]=name, [1]=weight, [2]=country of birth
    MEM.defn_sequence = MEM.preloadWithOwner(MEM.ANNE,{text: "definition sequence", nid:'7776892172829859353'}, "class",  MEM.sequence);
    
    //all insts of defn sequence will have a subject of protocol param, eg, they will sequence the order of the params in the protocol
    MEM.preLoadFullRelon(TYPE_HAS, MEM.defn_sequence, MEM.subject, MEM.protocol_param);

    //external protocols expect a definition sequence of class sequence
    MEM.preLoadFullRelon(TYPE_PARAM, MEM.external_protocol, MEM.defn_sequence, MEM.defn_sequence);

    //external protocols get their sequence neuron sorted as their set_value function
    MEM.external_protocol.handle=function($externlProtocolInst, $txid) {
        var seqRelon = MEMG.getRelonFromWithRToType($externlProtocolInst.outs, TYPE_HAS, MEM.defn_sequence, TO_TYPE_ANNE);
        if (seqRelon!=null) {
            var seqNOBJ = MEMG.getNeuronByTempNid(seqRelon.t);
            MEM.sequence.set_value(seqNOBJ);  //this will sort the sequence
        }
        var handler = MEMG.getRelonFromWithRToType($externlProtocolInst.outs, TYPE_HAS, MEM.protocol_handler, TO_TYPE_ANNE);
        if (handler!=null) {
            var handlerNOBJ = MEMG.getNeuronByTempNid(handler.t);
            handlerNOBJ.handle($txid, seqNOBJ);
        }
    }

    //these never actually got made, thus these nids are not good.

    HANDLER.json_handler = MEM.preloadWithOwner(MEM.ANNE,{text: "json handler", nid:'7775250413280208217'}, "class",  MEM.handler);
    HANDLER.json_object = MEM.preloadWithOwner(MEM.ANNE,{text: "json object", nid:'7778715200323828174'}, "class",  MEM.object);
    HANDLER.json_string = MEM.preloadWithOwner(MEM.ANNE,{text: "json string", nid:'7772990881420108491'}, "class",  MEM.string)
    HANDLER.json_boolean = MEM.preloadWithOwner(MEM.ANNE,{text: "json boolean", nid:'7779856355794103011'}, "class",  MEM.boolean)
    HANDLER.json_date = MEM.preloadWithOwner(MEM.ANNE,{text: "json date", nid:'7773511930066699925'}, "class",  MEM.date)
    HANDLER.json_unique_id = MEM.preloadWithOwner(MEM.ANNE,{text: "json unqique id", nid:'777828082708386020'}, "class",  MEM.ID)
    HANDLER.json_url = MEM.preloadWithOwner(MEM.ANNE,{text: "json url", nid:'7779622624493225622'}, "class",  MEM.url)
    HANDLER.json_integer = MEM.preloadWithOwner(MEM.ANNE,{text: "json integer", nid:'7770278105202396462'}, "class",  MEM.integer)
    HANDLER.json_blob = MEM.preloadWithOwner(MEM.ANNE,{text: "json blob", nid:'7777481949978186096'}, "class",  MEM.blob)
    HANDLER.json_mapper_instruct = MEM.preloadWithOwner(MEM.ANNE,{text: "json_mapper_instruct", nid:'7774893884114208359'}, "class",  MEM.class);
    HANDLER.target_ANNE_type = MEM.preloadWithOwner(MEM.ANNE,{text: "target ANNE type", nid:'7777061409751980420'}, "class",  HANDLER.json_mapper_instruct)
    HANDLER.target_ANNE_class = MEM.preloadWithOwner(MEM.ANNE,{text: "target ANNE class", nid:'7770936894749179393'}, "class",  HANDLER.json_mapper_instruct)
    HANDLER.target_ANNE_relon = MEM.preloadWithOwner(MEM.ANNE,{text: "target ANNE relon", nid:'7771056595422757266'}, "class",  HANDLER.json_mapper_instruct)
    HANDLER.target_ANNE_relon_reln = MEM.preloadWithOwner(MEM.ANNE,{text: "target_ANNE_relon_reln", nid:'7771427242546652011'}, "class",  HANDLER.json_mapper_instruct)
    HANDLER.target_ANNE_relon_to = MEM.preloadWithOwner(MEM.ANNE,{text: "target_ANNE_relon_to", nid:'7774996693954687390'}, "class",  HANDLER.json_mapper_instruct)
    HANDLER.target_ANNE_name_prop = MEM.preloadWithOwner(MEM.ANNE,{text: "target_ANNE_name_prop", nid:'7772734999169893878'}, "class",  HANDLER.json_mapper_instruct)
    HANDLER.target_ANNE_name_prop_obj = MEM.preloadWithOwner(MEM.ANNE,{text: "target_ANNE_name_prop_obj", nid:'7772837366359190563'}, "class",  HANDLER.json_mapper_instruct)
    HANDLER.target_ANNE_self_obj = MEM.preloadWithOwner(MEM.ANNE,{text: "target_ANNE_self_obj", nid:'7779575932224603440'}, "class",  HANDLER.json_mapper_instruct)
    HANDLER.target_ANNE_value_mapper = MEM.preloadWithOwner(MEM.ANNE,{text: "target_ANNE_value_mapper", nid:'7776047796377621219'}, "class",  HANDLER.json_mapper_instruct)
    HANDLER.target_ANNE_to_type = MEM.preloadWithOwner(MEM.ANNE,{text: "target_ANNE_to_type", nid:'7772799677381712865'}, "class",  HANDLER.json_mapper_instruct)
    HANDLER.target_ANNE_prepend_val = MEM.preloadWithOwner(MEM.ANNE,{text: "target_ANNE_prepend_val", nid:'7779035717561788114'}, "class",  HANDLER.json_mapper_instruct)
    HANDLER.target_ANNE_repeat_top = MEM.preloadWithOwner(MEM.ANNE,{text: "target_ANNE_repeat_top", nid:'7772332432887640062'}, "class",  HANDLER.json_mapper_instruct)
    HANDLER.target_ANNE_name_override = MEM.preloadWithOwner(MEM.ANNE,{text: "target_ANNE_name_override", nid:'7779490665992242363'}, "class",  HANDLER.json_mapper_instruct);
    HANDLER.iflastnull_trynext = MEM.preloadWithOwner(MEM.ANNE,{text: ":iflastnull_trynext", nid:'7779485596596330816'}, "class",  HANDLER.json_mapper_instruct);
    HANDLER.iflastnull_exit = MEM.preloadWithOwner(MEM.ANNE,{text: ":iflastnull_exit", nid:'7772381298092045676'}, "class",  HANDLER.json_mapper_instruct);
    MEM.calculated_value = MEM.preloadWithOwner(MEM.ANNE,{text: "calculated_value", nid:'7773206092203604463'}, "class",  HANDLER.json_mapper_instruct)
    

    /*** HANDLER.json_handler.handle ************************* xx/xx/2019 ***
    * The mapNOBJ is the type of object we are expecting to parse in the    *
    * JSON.  It is dynamically defined in its own neuron-circuit.  This     *
    * allows for dynamic handling of any future API data ---> with the end  *
    * outcome of it being encoded into ANNE and interconnected into the     *
    * graph.                                                                *
    ******************************************************** Mr.Scatman ****/
    HANDLER.json_handler.handle = function($mapNOBJ, $json, $newNOBJ=null, $nest=0, $neuronsToResolve) {
        var outs = $mapNOBJ.outs;
        var _neuronsToResolve = [];

        if ($neuronsToResolve!=null) _neuronsToResolve = $neuronsToResolve;

        function makeNewRelon($lastNewNOBJ, $ANNEmapPropNOBJ, $reln_nObj, $to_var, $to_neuron, $nest, $debugFromFunct) {
            if ($lastNewNOBJ==null) {
                console.log("(makeNewRelon) error, no lastNewNOBJ");
                return;
            }
            
            var to_type = "";
            if ($to_neuron!=null) {
                to_type = parseInt(TO_TYPE_ANNE);
            }
            else if ($ANNEmapPropNOBJ.temp_nid== MEM.calculated_value.temp_nid) {
                to_type=parseInt(TO_TYPE_STRING);
            }
            else { 
                //for all relons to non-neurons
                var TT_NOBJ = MEMG.getTerminatingType($ANNEmapPropNOBJ);
                if (TT_NOBJ!=null) to_type = TT_NOBJ.SYS_ENUM_VAL;
                else {
                    var to_type_relon  = MEMG.getRelonFromWithTempNid($reln_nObj.outs, HANDLER.target_ANNE_to_type.temp_nid);
                    if (to_type_relon!=null) {
                        to_type = parseInt(to_type_relon.to_s);
                    }
                    else {
                        var TT_NOBJ = MEMG.getTerminatingType($reln_nObj);
                        if (TT_NOBJ!=null) to_type = TT_NOBJ.SYS_ENUM_VAL;
                        else {
                            console.log("UNHANDLED RELN NOBJ --> UNKNOWN TO TYPE FOR makeNewReloN '"+$reln_nObj.text+"'");
                            return;
                        }
                    }
                }
            }
            var cmd = ANNE.BUNDLE.getCmdFromType(to_type, 'has');
            var reln = $ANNEmapPropNOBJ.nid;
            var reln_text = $ANNEmapPropNOBJ.text;
            var to_var =  $to_var;
            
            //we have a little work to do yet..  (dynamic relon reln based on value in relon_reln and relon_to)
            //we must find both of these values, to get the reln and the actual to,
            //otherwise, bail out.
            if ($ANNEmapPropNOBJ.temp_nid==MEM.calculated_value.temp_nid) {
                var reln_relon  = MEMG.getRelonFromWithTempNid($reln_nObj.outs, HANDLER.target_ANNE_relon_reln.temp_nid);
                var relnNOBJ = null;
                if (reln_relon!=null) relnNOBJ = MEMG.getNeuronByTempNid(reln_relon.t);
                var to_relon  = MEMG.getRelonFromWithTempNid($reln_nObj.outs, HANDLER.target_ANNE_relon_to.temp_nid);
                var toNOBJ = null;
                if (to_relon!=null) toNOBJ = MEMG.getNeuronByTempNid(reln_relon.t);
                if (relnNOBJ!=null && toNOBJ!=null) {
                    reln = relnNOBJ.nid;
                    reln_text = relnNOBJ.text;
                    to_var = $to_var[ toNOBJ.text.trim()];
                }
            }

            //override for ANNE mapper relon reln (dynamically set based on a mapper value)
            else if ($ANNEmapPropNOBJ.temp_nid==HANDLER.target_ANNE_value_mapper.temp_nid) {
                var cdx_val = "";
                var valtype = typeof $lastNewNOBJ.json;
                if (Array.isArray($lastNewNOBJ.json)==true) cdx_val = $lastNewNOBJ.json[0][ $ANNEmapPropNOBJ.text.trim() ];
                else if (valtype === 'object' || valtype === 'Object') cdx_val = $lastNewNOBJ.json[ $ANNEmapPropNOBJ.text.trim() ];
                var relonx = ANNE.MEM.getRelonFromTo($ANNEmapPropNOBJ.outs, TYPE_HAS, cdx_val);
                if (relonx!=null) {
                    var newMapper = MEMG.getNeuronByTempNid(relonx.t);
                    if (newMapper!=null) {
                        reln_text = newMapper.text;
                        reln = newMapper.reln;
                    }
                }
            }
            
            //no empty relons.
            if (to_var=="" && $to_neuron==null) return null;
            
            //prepend check, only applies to relon to_s values.
            if ($reln_nObj!=null) { 
                var prepend  = MEMG.getRelonFromWithTempNid($reln_nObj.outs,  HANDLER.target_ANNE_prepend_val.temp_nid);
                if (prepend!=null) to_var = prepend.to_s.trim()+to_var;
            }
            if (cmd=="has_number") to_var = parseFloat(to_var);

            return HANDLER.__makeNewRelon(TYPE_HAS, reln_text, reln, to_var, $to_neuron, cmd, $nest, $debugFromFunct, $lastNewNOBJ);
        }

        function handleArray($arr, $classNObj, $typeNObj, $nameProp="", $namePropObjRelon=null, $reln_nObj, $lastNewNOBJ=null, $nest=0) {
            for (var i=0; i < $arr.length; i++) { 
                var item = $arr[i];
                var valtype = typeof item;
                //console.log(item);
                //console.log(valtype);
                if (valtype === "object" || valtype === "Object") {
                    handleObj(item, $classNObj, $typeNObj, $nameProp, $namePropObjRelon, $reln_nObj, $lastNewNOBJ, $nest);
                }

                //the item is a simple prop (this was an array of values, not an array of objects)
                //we pass classNobj and type in case this thing is its own neuron (but not a nested object)
                //ie, category.  Category is not just a free-entry string in ANNE, it is a class of categories.
                else {
                    handleProp(item, $reln_nObj, $lastNewNOBJ, $classNObj, $typeNObj, $nest);
                }
            }
        }

        function handleObj($obj, $classNObj, $typeNObj, $nameProp, $namePropObjRelon=null, $reln_nObj, $lastNewNOBJ=null, $nest=0) {
            var ANNEMAPRELON  = MEMG.getRelonFromWithTempNid($reln_nObj.outs, HANDLER.target_ANNE_relon.temp_nid);
            var ANNEmapPropNOBJ = null;
            if (ANNEMAPRELON!=null) {
                ANNEmapPropNOBJ = MEMG.getNeuronByTempNid(ANNEMAPRELON.t);
            }
            var newNOBJ = null;

            /* we need to create a new neuron ... */
            if ($classNObj!=null && $typeNObj!=null && $namePropObjRelon!=null) {
                if ($nameProp=="") {
                    console.log("handleObj - we can't create a new neuorn without a name prop");
                    return;
                }
                else  {
                    var text = "";
                    var namePropNOBJ = MEMG.getNeuronByTempNid($namePropObjRelon.t);
                    if (namePropNOBJ!=null && $nameProp!="" && namePropNOBJ.temp_nid==HANDLER.target_ANNE_self_obj.temp_nid) {
                        text = $obj[$nameProp];
                    }
                    else {
                        var namePropStr = namePropNOBJ.text.trim();
                        var namePropObj = $obj[namePropStr];
                        if (namePropObj!=undefined && namePropObj!=null) text = namePropObj[$nameProp];
                    }

                    if (text!="") {    
                        //console.log("new neuron name will be : "+text);
                        newNOBJ = HANDLER.__makeNewHandlerNeuron($typeNObj, text, $classNObj, $nest, true, $obj, _neuronsToResolve);
                    }
                    else {
                        console.log("cannot ascertain name of new neuron.");
                        return;
                    }
                }
            }

            //we need to create a new relon..
            if (ANNEmapPropNOBJ!=null) {
                if (newNOBJ!=null) {
                    makeNewRelon($lastNewNOBJ, ANNEmapPropNOBJ,  $reln_nObj, "", newNOBJ,  $nest, 'handleObj');
                }

                else {
                    if ($nameProp!="") var obj =$obj[$nameProp];
                    else var obj = $obj;
                    makeNewRelon($lastNewNOBJ, ANNEmapPropNOBJ,  $reln_nObj, obj, null,  $nest, 'handleObj');
                }
            }

            //this is an object, with properties, we need to recurse on it.
            if (newNOBJ!=null) {
                HANDLER.json_handler.handle($reln_nObj, $obj, newNOBJ, $nest+1, _neuronsToResolve);
            }
            else {
                HANDLER.json_handler.handle($reln_nObj, $obj, $lastNewNOBJ, $nest+1, _neuronsToResolve);
            }
        }

        //a basic prop that is a string, bool, number, float, int etc.
        //it is not an object and not an array.
        //it should have an ANNEMAPRELON, if not it is an error (do nothing)
        function handleProp($prop_value, $reln_nObj, $lastNewNOBJ=null, $classNObj=null, $typeNObj=null, $nest=0) {
            var ANNEMAPRELON  = MEMG.getRelonFromWithTempNid($reln_nObj.outs, HANDLER.target_ANNE_relon.temp_nid);
            var ANNEmapPropNOBJ = null;

            if ($classNObj!=null && ($classNObj.nid==MEM.gender.nid || $classNObj.nid==MEM.placeOfBirth.nid)) {
                console.warn("DEBUG HERE");
            }

            if (ANNEMAPRELON!=null) ANNEmapPropNOBJ = MEMG.getNeuronByTempNid(ANNEMAPRELON.t);
            if (ANNEmapPropNOBJ!=null) {
               var relon = null;

                //an array of props -- that did not generate a new class, only non TO_TYPE_ANNE relons..
                if (Array.isArray($prop_value)==true) {
                    var valueMapRelon  = MEMG.getRelonFromWithTempNid($reln_nObj.outs, HANDLER.target_ANNE_value_mapper.temp_nid);
                            
                    for (var p=0; p < $prop_value.length; p++) {
                        var prop_val = $prop_value[p];
                        
                        if (valueMapRelon==null && $classNObj!=null && $typeNObj!=null) {
                            to_neuron = HANDLER.__makeNewHandlerNeuron($typeNObj, prop_value, $classNObj, $nest+1, true, $prop_value, _neuronsToResolve);
                            relon = makeNewRelon($lastNewNOBJ, ANNEmapPropNOBJ, $reln_nObj,  prop_value, to_neuron, $nest, 'handleProp');
                            return;
                        }

                        relon = makeNewRelon($lastNewNOBJ, ANNEmapPropNOBJ, $reln_nObj, prop_val, null, $nest, 'handleProp');
                        if (relon!=null && $classNObj!=null && $typeNObj!=null) {
                            var to_neuron = null;
                            if (valueMapRelon!=null && valueMapRelon.to_s!=null && valueMapRelon.type==TYPE_BE) {
                                var valueMapNOBJ = ANNE.MEMG.getNeuronByNid(valueMapRelon.to_s);
                                //if we match this val to the to of any relns in the valuemapper, the reln is the resolution!
                                var reln = ANNE.MEM.getRelnFromTo(valueMapNOBJ.outs, TYPE_HAS, prop_val);
                                if (reln!="") {
                                    relon.cmd = relon.cmd.split("_")[0]; //strip the special part off the cmd, this is a real anne address now.
                                    relon.to_s = reln;
                                    relon.to_type = TO_TYPE_ANNE;
                                    relon.to_neuron = null;
                                }
                                else  to_neuron = HANDLER.__makeNewHandlerNeuron($typeNObj, prop_val, $classNObj, $nest+1, false, $prop_value, _neuronsToResolve);
                            }
                            else {
                                 to_neuron = HANDLER.__makeNewHandlerNeuron($typeNObj, prop_val, $classNObj, $nest+1, false, $prop_value, _neuronsToResolve);
                            }

                            if (to_neuron!=null) relon.to_neuron=to_neuron;
                        }
                    }
                    return;
                }

                else {
                    var valueMapRelon  = MEMG.getRelonFromWithTempNid($reln_nObj.outs, HANDLER.target_ANNE_value_mapper.temp_nid);
                        
                    //for basic values with no mapper that just simply create a new neuron (no relons)
                    //resolution will have to do its magic
                    if (valueMapRelon==null && $classNObj!=null && $typeNObj!=null) {
                        to_neuron = HANDLER.__makeNewHandlerNeuron($typeNObj, $prop_value, $classNObj, $nest+1, true, $prop_value, _neuronsToResolve);
                        relon = makeNewRelon($lastNewNOBJ, ANNEmapPropNOBJ, $reln_nObj,  $prop_value, to_neuron, $nest, 'handleProp');
                        return;
                    }

                    relon = makeNewRelon($lastNewNOBJ, ANNEmapPropNOBJ, $reln_nObj,  $prop_value, null, $nest, 'handleProp');
                    if (relon!=null && $classNObj!=null && $typeNObj!=null) {
                        var to_neuron = null;
                        if (valueMapRelon!=null && valueMapRelon.to_s!=null && valueMapRelon.type==TYPE_BE) {
                            var valueMapNOBJ = ANNE.MEMG.getNeuronByNid(valueMapRelon.to_s);
                            //if we match this val to the to of any relns in the valuemapper, the reln is the resolution!
                            var reln = ANNE.MEM.getRelnFromTo(valueMapNOBJ.outs, TYPE_HAS, $prop_value);
                            if (reln!="") {
                                relon.cmd = relon.cmd.split("_")[0]; //strip the special part off the cmd, this is a real anne address now.
                                relon.to_s = reln;
                                relon.to_type = TO_TYPE_ANNE;
                                relon.to_neuron = null;
                            }
                            else  to_neuron = HANDLER.__makeNewHandlerNeuron($typeNObj, prop_val, $classNObj, $nest+1, false, $prop_value, _neuronsToResolve);
                        }       
                        
                        if (to_neuron!=null) relon.to_neuron=to_neuron;
                    }
                }
            }
            else {
                console.log("(handleProp) ERROR, no maprelon found for prop_value:"+$prop_value +" reln text:"+ $reln_nObj.text);
            }
        }


        function handleHas($rOBJ, $activeNOBJ=null, $nest=0) {

            if ($rOBJ.reln==MEM_MOVIEDB.mdb_cast_order.temp_nid || $rOBJ.to_s=="52fe44e6c3a36847f80b09cf") {
                console.warn("debug here");
            }

            var reln_nOBJ = MEMG.getNeuronByTempNid($rOBJ.r);        
            var classRelon = MEMG.getRelonFromWithTempNid(reln_nOBJ.outs, HANDLER.target_ANNE_class.temp_nid);
            var classNObj = null;
            if (classRelon!=null) {
                classNObj = MEMG.getNeuronByTempNid(classRelon.t);
                if (classNObj.emulated==true) classNObj=null;
            }
            var typeRelon  = MEMG.getRelonFromWithTempNid(reln_nOBJ.outs, HANDLER.target_ANNE_type.temp_nid);
            var typeNObj = null;
            if (typeRelon!=null) typeNObj = MEMG.getNeuronByTempNid(typeRelon.t);
            var nameProp = "";
            var namePropRelon = MEMG.getRelonFromWithTempNid(reln_nOBJ.outs, HANDLER.target_ANNE_name_prop.temp_nid);
            if (namePropRelon!=null) nameProp = namePropRelon.to_s.trim();
            var namePropObjRelon = MEMG.getRelonFromWithTempNid(reln_nOBJ.outs, HANDLER.target_ANNE_name_prop_obj.temp_nid);
            var prop = reln_nOBJ.text.trim();

            var someValOrObj = $json[prop];
            if (someValOrObj==undefined) {
                return;
            }

            var valtype = typeof someValOrObj;

            //this will update the class of the neuron being created to a dynamic value set on a 
            //mapper that pts to an equal level prop (mapper) and compares it to relons inside of the mapper neuron
            if (classNObj!=null) {
                if (classNObj.bes[0]==HANDLER.target_ANNE_value_mapper.temp_nid) {
                    var cdx_val = "";
                    if (Array.isArray(someValOrObj)==true) cdx_val = someValOrObj[0][ classNObj.text.trim() ];
                    else if (valtype === 'object' || valtype === 'Object') cdx_val = someValOrObj[ classNObj.text.trim() ];
                    
                    if (cdx_val) {
                        var relon = ANNE.MEM.getRelonFromTo(classNObj.outs, TYPE_HAS, cdx_val);
                        if (relon!=null) {
                            classRelon = relon;
                            classNObj = MEMG.getNeuronByTempNid(relon.r);
                            if (classNObj.emulated==true) classNObj = null; //safety override
                        }
                    }
                }
            }


            //a nested object..
            if ($rOBJ.t == HANDLER.json_object.temp_nid) {
                //console.warn(someValOrObj);
                //we wil recurse on any objects within this array (once we process the array)
                if (Array.isArray(someValOrObj)==true) {
                    //console.log("this is an array at nest "+$nest);
                    handleArray(someValOrObj, classNObj, typeNObj, nameProp, namePropObjRelon, reln_nOBJ, $activeNOBJ, $nest);
                }
                else if (valtype === 'object' || valtype === 'Object') {
                    //console.log("this is an object at nest "+$nest);
                    handleObj(someValOrObj, classNObj, typeNObj, nameProp, namePropObjRelon, reln_nOBJ, $activeNOBJ, $nest);
                }

                //a string, mapped to a new neuron (resolution will have to do its magic)
                else {
                    handleProp(someValOrObj, reln_nOBJ, $activeNOBJ, classNObj, typeNObj, $nest);
                }
               
            }

            //no recursing on these, we are at the end of the line.
            else {
                handleProp(someValOrObj, reln_nOBJ, $activeNOBJ, classNObj, typeNObj, $nest);
            }
        }

        //************************************************//
        //******************* MAIN LOOP ******************//
        //************************************************//
        for (var r=0; r <outs.length; r++) {
            var rOBJ = outs[r];
            if (rOBJ.type!=TYPE_HAS) continue;
            if (rOBJ.r == MEM.handler.temp_nid) continue;
            
            //TALK.printRelon(rOBJ);
            var RNOBJ = MEMG.getNeuronByTempNid(rOBJ.r);

            //we don't need to process mappers in the handleHas
            //we are looking for real values created by the API cicruit in question (ie, movies)
            
            //special outter handling case -- we want to create a relon from this thing to a value
            //in the original upper neuron  (nested neuron requires a prop value from the parent obj).
            if (RNOBJ.nid == HANDLER.target_ANNE_repeat_top.nid) {
                var repNOBJ = ANNE.MEMG.getNeuronByTempNid(rOBJ.t);
                if (repNOBJ==null) continue;
                var rProp = repNOBJ.text.trim();
                var repeat_val = _neuronsToResolve[0].json[rProp];
                
                //get the target relon
                //for now, the reater is just a simple value.
                var ANNEMAPRELON  = ANNE.MEMG.getRelonFromWithTempNid(repNOBJ.outs, HANDLER.target_ANNE_relon.temp_nid);
                if (ANNEMAPRELON==null) continue;
                var toNOBJ=ANNE.MEMG.getNeuronByTempNid(ANNEMAPRELON.t);
                if (toNOBJ==null) continue;
                makeNewRelon($newNOBJ, toNOBJ, toNOBJ, repeat_val, null, 0, 'repeater top');
                continue;
            }

            if (RNOBJ!=null && RNOBJ.bes[0]==HANDLER.json_mapper_instruct.temp_nid) continue;
            
            handleHas(rOBJ, $newNOBJ, $nest);
        }

        if ($nest==0) return _neuronsToResolve;
    }  

    return true;
}

HANDLER.__makeNewRelon=function($type, $reln_text, $reln, $to_var, $to_neuron, $cmd, $nest, $debugFromFunct, $lastNewNOBJ) {      
    var relon = new Object();
    relon.type = $type;
    relon.relon_text = $reln_text;
    relon.reln = $reln;
    relon.to_s = $to_var;
    relon.to_neuron = $to_neuron;
    relon.cmd = $cmd;
    relon.nest = $nest;
    relon.debug = $debugFromFunct;
    relon.from_neuron = $lastNewNOBJ;
    relon.handler_relon = true;

    var outs = $lastNewNOBJ.relons;
    if (outs.length==0) {
        $lastNewNOBJ.relons.push(relon);
        return relon;
    }
    else {
        for (var r=0; r < outs.length; r++) {
            var rel = outs[r];
            if (rel.type==$type && rel.reln==relon.reln && rel.to_s==relon.to_s
                    && rel.to_neuron==relon.to_neuron && relon.cmd==$cmd && relon.nest==$nest) {
                        console.log("skipping dupe relons to "+relon.to_s);
                return null;
            }
        }
    }
    $lastNewNOBJ.relons.push(relon);
    return relon;
}

HANDLER.__makeNewHandlerNeuron=function($typeNObj, $text, $classNObj, $nest=0, $bPUSH=true, $json={}, $ntr=[]) {
    var neuron = new Object();
    neuron.type = $typeNObj.text;
    neuron.text = $text;
    neuron.parentClass = $classNObj;
    neuron.relons = [];
    neuron.nest = $nest;
    neuron.resolved = false;
    neuron.resolved_nObj = null;
    neuron.json = $json;
    neuron.temp_nid = -1;   //as to not confuse
    if ($bPUSH==true) $ntr.push(neuron);
    return neuron;
}

/*** getProp ********************************************* 0x/xx/2019 ***
******************************************************** Mr.Scatman ****/
HANDLER.getProp = function ($thing, $rNObj) {
    if ($rNObj==null) {
        console.log("huh??");
        return null;
    }
    var relons = $thing.relons;
    for (var r=0; r < relons.length; r++) {
        var rObj = relons[r];
        if (rObj.reln==$rNObj.nid) {
            if (rObj.to_neuron==null) return rObj.to_s;
            else return rObj.to_neuron;
        }
    }
    return null;
}


/*** getProps ******************************************** 0x/xx/2019 ***
******************************************************** Mr.Scatman ****/
HANDLER.getProps = function ($book, $rNObj) {
    var relons = $book.relons;
    var objs = [];
    for (var r=0; r < relons.length; r++) {
        var rObj = relons[r];
        if (rObj.reln==$rNObj.nid) {
            if (rObj.to_neuron==null) objs.push(rObj.to_s);
            else objs.push(rObj.to_neuron);
        }
    }
    return objs;
}


/*** resolveAPIGeneralizedResult ************************* 0x/xx/2019 ***
******************************************************** Mr.Scatman ****/
HANDLER.resolveAPIGeneralizedResult = async ($thing, $primaryIdNOBJ, $WVS=[], $resolvables=[], $mustBeResolved=[], $chaseables=[], $bRETURNIFTOPRESOLVED=true, $bLOOKUPTOPRELONS=false)=> {
    var id = HANDLER.getProp($thing, $primaryIdNOBJ);    //get primary key id.
    $thing.ezid = id;

    if ($thing.parentClass.postParseFix!=undefined) $thing.parentClass.postParseFix($thing);

    //this will look up local memory and then try mongo atlas.
    var nObj_thing = null;
    if (id!=null) nObj_thing = await MEM.getNeuronByUniqueRelon($thing.parentClass, $primaryIdNOBJ, id, $WVS, []);
    if (nObj_thing!=null) {
        $thing.resolved = true;
        $thing.resolved_nObj = nObj_thing;
        $thing.resolved_nObj.ezid = id;
        if ($bRETURNIFTOPRESOLVED==true) return;
    }
    else {
        $thing.resolved = false;
        $thing.resolved_nObj = false;
    }

    //resolve innerds
    var relons = $thing.relons;
    if (relons!=undefined  && relons.length) await HANDLER.resolveAPIGeneralizedInnerValues(relons, $WVS, $resolvables, $mustBeResolved, $chaseables, nObj_thing, $bLOOKUPTOPRELONS);
    return id;
}


/*** resolveAPIGeneralizedInnerValues ******************** 0x/xx/2019 ***
* here we to resolve all nested neurons from the API parse.             *
* Depending on the run and use case, it may not be necessary to try to  *
* resolve some items.   In MM if a person is not resolved then their    *
* movie credits will not be resolved either -- unless they were         *
* sponsored from a super movie sponsore which has 1 of them already.    *
* if something is to be resolved that aid should be passed into the  *
* resolvables.                                                          *
*                                                                       *
* **PARAMS**                                                            *
*                                                                       *
* relons - the outs of this neuron we are trying to resolve.            *
* resolvables - classes of neurons we want to try to resolve.           *
* mustBeResolved - classes of neurons we MUST resolve.  These would be  *
*   classes that we expect to have all options in memory for -- and we  *
*   are mapping, not creating new things of.                            *
* chaseables - classes we are ok chasing mongo for.  These are costly   *
*   lookups, so use them wisely.                                        *
*                                                                       *
******************************************************** Mr.Scatman ****/
HANDLER.resolveAPIGeneralizedInnerValues=async($relons, $WVS, $resolvables=[], $mustBeResolved=[], $chaseables=[], $nObj=null, $bLOOKUPTOPRELONS=false)=> {
    console.warn("resolveAPIGeneralizedInnerValues");
    for (var r=0; r < $relons.length; r++) {
        var rObj = $relons[r];
        if (rObj.to_neuron!=null) {
            var toNObj = rObj.to_neuron;
            var toNObjResolved = null;
            var parentClass = toNObj.parentClass;

            if (parentClass.postParseFix!=undefined) {
                parentClass.postParseFix(toNObj);
            }

            
            //if this is not a resolvable neuron, skip it.
            //this means we wanted this in the relons array of the handler neuron but we do not want
            //to actually resolve it here.
            if ($resolvables.length && $resolvables.indexOf(parentClass.nid)==-1) continue;

            //try a custom parent class resolver (if it exists)
            if (parentClass.resolve!=undefined && typeof parentClass.resolve=="function") {
                var neurons_needed = [];
                var resolverNOBJ = await parentClass.resolve(toNObj, neurons_needed, $chaseables);
                if (resolverNOBJ==null || (resolverNOBJ.resolved==false)) {
                    toNObjResolved = null;
                    rObj.to_neuron = resolverNOBJ;  //this tells us we have to create this unresolved NOBJ (new)
                }
                else toNObjResolved = resolverNOBJ;
            }
            else if (toNObj!=null && toNObj.resolved!=undefined && toNObj.resolved==true) {
                toNObjResolved = toNObj.resolved_nObj;
            }
            
            else {
                //see if this has a primary key relon associated with it.
                //something like country has a primary prop of MEM.iso_3166_1_alpha2
                //something like language has a primary prop of iso_639_1.
                var pkRelon = ANNE.MEMG.getRelonFromWithRToType(parentClass.outs, TYPE_HAS, MEM.primary_prop.nid, TO_TYPE_ANNE);
                if (pkRelon!=null) {
                    var pkNOBJ = ANNE.MEMG.getNeuronByTempNid(pkRelon.t);
                    var thisPk = HANDLER.getProp(toNObj, pkNOBJ);    //get primary key id.
                    toNObj.ezid=thisPk;
                    toNObjResolved = MEM_UDF.getPkResolvedChild(parentClass, thisPk);
                    if (toNObjResolved==null) {
                        toNObjResolved = await MEM.getNeuronByUniqueRelon(parentClass, pkNOBJ, thisPk, $WVS, $chaseables);
                    }
                    if (toNObjResolved!=null) {
                        toNObjResolved.ezid=thisPk;
                        MEM_UDF.addPkResolvedChild(parentClass, toNObjResolved, thisPk);
                    }
                }

                //try basic parent class w/ TEXT lookup.
                else {
                    var to_s = rObj.to_s;
                    if (rObj.to_neuron!=null) to_s = rObj.to_neuron.text;
                    toNObjResolved = await ANNE.CHASE.getChildWithParentAndText(parentClass, to_s, $WVS, $chaseables);
                }
            }

            if (toNObjResolved!=null) {
                if ($nObj==null) {
                    console.log("%c <<resolved neuron!>> "+toNObjResolved.text+" (be:"+parentClass.text+") '"+toNObj.text+"'", 'background: green; color: white');
                }
                else rObj.active=false;  //skip this on re-pricing for new data.

                toNObj.resolved=true;
                rObj.to_s = toNObjResolved.nid;
                rObj.to_type = TO_TYPE_ANNE;
                rObj.cmd = rObj.cmd.split("_")[0];
                rObj.to_neuron.resolved=true;
                rObj.to_neuron.resolved_nObj = toNObjResolved;
                
                //if this is a redo, we stil need to check if we burned this relon already
                //(to the known resolved thing thing)
                if ($bLOOKUPTOPRELONS==true && $nObj!=null) {
                    var sch = parseInt(rObj.type);
                    var to_type = MEM.getTypeByCmd(rObj.cmd);
                    to_type = parseInt(to_type);
                    var to_s = rObj.to_s;
                    if (rObj.to_s==undefined) {
                        console.log("error, to_s not found");
                        rObj.active = false;
                        continue;
                    }
                    if (to_type==TO_TYPE_BLOB || to_s.length > 128) {
                        to_s = ANNE.other.sha256(to_s);
                        var mem_relon = MEMG.getRelonFromWithToHash($nObj.outs, sch, rObj.reln, to_s);
                    }
                    else var mem_relon = MEMG.getRelonFromWithTo($nObj.outs, sch, rObj.reln, to_s);

                    if (mem_relon!=null) rObj.active=false;
                    else {
                        console.log("we did not find reln to "+ toNObjResolved.text+" ("+rObj.to_s+")");
                        rObj.active=true;
                        var rNOBJ = MEMG.getNeuronByNid(rObj.reln);
                        rObj.fix_msg= rNOBJ.text+"<b>&nbsp;edge to&nbsp;</b> "+toNObjResolved.text;
                    }
                }
            }
            else {
                if ($mustBeResolved.length && $mustBeResolved.indexOf(parentClass.nid)!=-1) {
                    rObj.to_neuron=null;
                    rObj.active=false;
                    console.log("%c **DID NOT RESOLVE A MUST RESOLVE** '"+toNObj.text+"'", 'background: red; color: white');
                    rObj.fix_msg="<span class='text-danger'>(trouble) We must resolve "+toNObj.text+" but didn't. Skipping</span>";
                }

                if ($nObj!=null) {
                    console.log("%c **DID NOT RESOLVE** '"+toNObj.text+"'", 'background: red; color: white');
                    console.log(toNObj);
                    rObj.fix_msg="<b>new neuron</b>&nbsp;"+toNObj.text;
                }
            }
        }

        //use local memory.  in this mode we have already pulled the existing has relons.
        //if this is one of them, mark it as active=false so we don't do it again.
        else if ($bLOOKUPTOPRELONS==true && $nObj!=null) {
            var sch = parseInt(rObj.type);
            var to_type = MEM.getTypeByCmd(rObj.cmd);
            to_type = parseInt(to_type);
            var to_s = rObj.to_s;
            if (to_type==TO_TYPE_BLOB || rObj.to_s.length > 128) {
                to_s = ANNE.other.sha256(rObj.to_s);
                var mem_relon = MEMG.getRelonFromWithToHash($nObj.outs, sch, rObj.reln, to_s)
            }
            else var mem_relon = MEMG.getRelonFromWithTo($nObj.outs, sch, rObj.reln, to_s);
            
            if (mem_relon!=null) {
                rObj.resolved_relon = mem_relon;
                rObj.active=false;
            }
            else {
                rObj.resolved_relon = null;
                console.log("%c **DID NOT FIND RELON TO** '"+rObj.to_s+"'", 'background: red; color: white');
                rObj.active=true;
                var rNOBJ = MEMG.getNeuronByNid(rObj.reln);
                rObj.fix_msg = rNOBJ.text+"&nbsp;<b>edge to</b>&nbsp;"+rObj.to_s;
            }
        }
    }
}


/*** getTextListFromHandlerNObjs ************************* 0x/xx/2019 ***
* We want to generate an array of text strings from the handler nobjs   *
* of a specific be class.  We will use these to do an optimized lookup. *
* SEE: CHASE.queryTextAndBes  for a typical use-case.                  *
* If bMARKRESOLVED is true we want to look these up in local mem first  *
* and mark them as resolved if found.  We do not need to chase them.    *
******************************************************** Mr.Scatman ****/
HANDLER.getTextListFromHandlerNObjs=function($unresolvedHOBJS, $beNOBJ, $bMARKRESOLVED=true, $WVS=[]) {
    var matches = [];
    var be_temp_nid = $beNOBJ.temp_nid;
    for (var n=0; n  < $unresolvedHOBJS.length; n++) {
        var uhOBJ = $unresolvedHOBJS[n];
        if (uhOBJ.parentClass.temp_nid==be_temp_nid) {
            if ($bMARKRESOLVED==false) matches.push(uhOBJ.text);
            else {
                var nObjxx = MEM.getChildByTextAndParent($beNOBJ, uhOBJ.text, $WVS);
                if (nObjxx!=null) {
                    uhOBJ.resolved_nObj = nObjxx;
                    uhOBJ.resolved=true;
                }
                else matches.push(uhOBJ.text);
            }
        }
    }
    return matches;
}


/*** getTextListFromNestedHandlerNObjs ******************* 0x/xx/2020 ***
* get credit/crew hneurons -- and extract a relon by reln from their    *
* relons..                                                              *
******************************************************** Mr.Scatman ****/
HANDLER.getIDistFromNestedHandlerNObjs=function($unresolvedHOBJS, $class_nids, $target_reln) {
    var matches = [];
    for (var n=0; n  < $unresolvedHOBJS.length; n++) {
        var uhOBJ = $unresolvedHOBJS[n];
        if ($class_nids.indexOf(uhOBJ.parentClass.temp_nid)!=-1) {
            for (var r=0; r < uhOBJ.relons.length; r++) {
                var rObj = uhOBJ.relons[r];
                if (rObj.reln==$target_reln) {
                    if (matches.indexOf(rObj.to_s)==-1) matches.push(uhOBJ.to_s); 
                }
            }
        }
    }
    return matches;
}